import time
from multiprocessing import Manager as MPManager
from multiprocessing import Process

import pytest

from goofi.connection import Connection
from goofi.message import Message, MessageType


@pytest.mark.parametrize("backend", Connection.get_backends().keys())
def test_abstract_create(backend):
    try:
        mp_manager = MPManager()
        Connection.set_backend("mp", mp_manager)
    except AssertionError:
        # connection backend is already set
        pass

    with pytest.raises(TypeError):
        Connection._create()


@pytest.mark.parametrize("backend", Connection.get_backends().keys())
def test_create(backend):
    try:
        mp_manager = MPManager()
        Connection.set_backend("mp", mp_manager)
    except AssertionError:
        # connection backend is already set
        pass

    conn1, conn2 = Connection.create()
    assert conn1 is not None and conn2 is not None, f"{backend.__name__}.create() returned None"


@pytest.mark.parametrize("backend", Connection.get_backends().keys())
def test_super_init(backend):
    try:
        mp_manager = MPManager()
        Connection.set_backend("mp", mp_manager)
    except AssertionError:
        # connection backend is already set
        pass

    conn1, conn2 = Connection.create()
    assert hasattr(conn1, "_id"), f"{backend.__name__} does not call super().__init__()."
    assert hasattr(conn2, "_id"), f"{backend.__name__} does not call super().__init__()."


@pytest.mark.parametrize("obj", [1, "test", None, [1, 2, 3], {"a": 1}, Message(MessageType.PING, {})])
@pytest.mark.parametrize("backend", Connection.get_backends().keys())
def test_send_recv(obj, backend):
    try:
        mp_manager = MPManager()
        Connection.set_backend("mp", mp_manager)
    except AssertionError:
        # connection backend is already set
        pass

    conn1, conn2 = Connection.create()
    conn1.send(obj)
    assert conn2.recv() == obj, "Connection.send() and Connection.recv() didn't work"


@pytest.mark.parametrize("backend", Connection.get_backends().keys())
def test_close(backend):
    try:
        mp_manager = MPManager()
        Connection.set_backend("mp", mp_manager)
    except AssertionError:
        # connection backend is already set
        pass

    conn1, conn2 = Connection.create()

    # make sure the connection is not closed
    conn1.send(1)
    assert conn2.recv() == 1, "Connection.send() and Connection.recv() didn't work"

    conn1.close()

    # closing multiple times should not raise an error
    conn1.close()

    with pytest.raises(ConnectionError):
        conn1.send(1)
    with pytest.raises(ConnectionError):
        conn1.recv()

    conn2.close()
    conn2.close()


@pytest.mark.parametrize("backend", Connection.get_backends().keys())
def test_try_send(backend):
    try:
        mp_manager = MPManager()
        Connection.set_backend("mp", mp_manager)
    except AssertionError:
        # connection backend is already set
        pass

    conn1, conn2 = Connection.create()

    assert conn2.try_send(1), "Connection.try_send() should return True on non-empty connection"
    assert conn1.recv() == 1, "Connection.send() and Connection.recv() didn't work"

    conn1.close()
    conn2.close()


@pytest.mark.parametrize("backend", Connection.get_backends().keys())
def test_multiproc(backend):
    try:
        mp_manager = MPManager()
        Connection.set_backend("mp", mp_manager)
    except AssertionError:
        # connection backend is already set
        pass

    def _send(conn):
        conn.send(1)
        # wait for the other process to receive the message before closing the connection
        time.sleep(0.1)

    conn1, conn2 = Connection.create()
    p = Process(target=_send, args=(conn1,))
    p.start()

    assert conn2.recv() == 1, "Connection.send() and Connection.recv() didn't work"
    p.join()



import numpy as np

from goofi.data import Data, DataType
from goofi.node import Node
from goofi.params import IntParam


class FOOOFaperiodic(Node):
    def config_input_slots():
        return {"psd_data": DataType.ARRAY}

    def config_output_slots():
        return {
            "offset": DataType.ARRAY,
            "exponent": DataType.ARRAY,
            "cf_peaks": DataType.ARRAY,
            "cleaned_psd": DataType.ARRAY,
        }

    def config_params():
        return {
            "fooof": {
                "max_n_peaks": IntParam(5, 1, 20, doc="The maximum number of peaks to fit."),
            }
        }

    def setup(self):
        from fooof import FOOOF

        self.FOOOF = FOOOF

    def process(self, psd_data: Data):
        if psd_data is None or psd_data.data is None:
            return None

        # Create FOOOF object & set its parameters
        fm = self.FOOOF(max_n_peaks=self.params["fooof"]["max_n_peaks"].value)

        # Extract the PSD and freqs from the Data object
        psd = psd_data.data

        try:
            if psd_data.data.ndim == 1:
                freqs = np.array(psd_data.meta["channels"]["dim0"])
            else:  # if 2D
                freqs = np.array(psd_data.meta["channels"]["dim1"])
        except KeyError:
            raise ValueError("No frequency information. Make sure to pass a power spectrum with frequency information.")

        # Fit FOOOF model
        fm.fit(freqs, psd)

        # Extract aperiodic component and peak parameters
        aperiodic_params = fm.get_params("aperiodic_params")
        peak_params = fm.get_params("peak_params")
        cleaned_psd = fm._spectrum_peak_rm

        # Extracting specific parameters
        offset = aperiodic_params[0]  # Extracting the offset
        exponent = aperiodic_params[-1]  # Extracting the exponent (if there's no knee value)

        # Getting all center frequencies of detected peaks
        cf_peaks = peak_params[:, 0] if peak_params.size else np.array([])

        return {
            "offset": (offset, {}),
            "exponent": (exponent, {}),
            "cf_peaks": (cf_peaks, {}),
            "cleaned_psd": (cleaned_psd, {}),
        }



import numpy as np

from goofi.data import Data, DataType
from goofi.node import Node


class Select(Node):
    def config_input_slots():
        return {"data": DataType.ARRAY}

    def config_output_slots():
        return {"out": DataType.ARRAY}

    def config_params():
        return {"select": {"axis": 0, "include": "", "exclude": ""}}

    def setup(self):
        from mne import pick_channels

        self.pick_channels = pick_channels

    def process(self, data: Data):
        if data is None:
            return None

        axis = self.params.select.axis.value
        if axis < 0:
            axis = data.data.ndim + axis

        include = self.params.select.include.value.split(",") or []
        include = [ch.strip() for ch in include if len(ch.strip()) > 0]
        exclude = self.params.select.exclude.value.split(",") or []
        exclude = [ch.strip() for ch in exclude if len(ch.strip()) > 0]

        if f"dim{axis}" in data.meta["channels"]:
            # use channel names from metadata
            chs = data.meta["channels"][f"dim{axis}"]
            idxs = self.pick_channels(chs, include=include, exclude=exclude, ordered=False)
        else:
            # no channel names for this axis, use indices
            chs = [str(i) for i in range(data.data.shape[axis])]

            if include and ":" in include[0]:
                # slice notation
                if len(include) > 1:
                    raise ValueError("Only one slice can be selected at a time.")
                if len(exclude) > 0:
                    raise ValueError("Excluding channels is not supported with slice notation.")
                slice_parts = include[0].split(":")
                start, stop, step = (int(x) if x else None for x in slice_parts + [None] * (3 - len(slice_parts)))
                idxs = chs[slice(start, stop, step)]
            else:
                if len(include) == 0:
                    # include all channels
                    include = chs

                # convert to indices and shift negative indices
                include = [int(ch) if int(ch) >= 0 else data.data.shape[axis] + int(ch) for ch in include]
                exclude = [int(ch) if int(ch) >= 0 else data.data.shape[axis] + int(ch) for ch in exclude]
                idxs = [i for i in range(len(chs)) if i in include and i not in exclude]

        if len(idxs) == 0:
            raise ValueError("No channels matched the selection.")

        # select channels from data
        selected = np.take(data.data, idxs, axis=axis)

        # update channel names if present
        if f"dim{axis}" in data.meta["channels"]:
            data.meta["channels"][f"dim{axis}"] = [ch for i, ch in enumerate(data.meta["channels"][f"dim{axis}"]) if i in idxs]

        if len(idxs) == 1:
            # remove axis if only one channel is selected
            selected = np.squeeze(selected, axis=axis)
            if f"dim{axis}" in data.meta["channels"]:
                del data.meta["channels"][f"dim{axis}"]

            for i in range(axis, selected.ndim + 1):
                if f"dim{i+1}" in data.meta["channels"]:
                    data.meta["channels"][f"dim{i}"] = data.meta["channels"][f"dim{i+1}"]
                    del data.meta["channels"][f"dim{i+1}"]

        return {"out": (selected, data.meta)}



import numpy as np
from goofi.node import Node
from goofi.data import DataType, Data
import colorsys


class RGBtoHSV(Node):
    def config_input_slots():
        return {"rgb_image": DataType.ARRAY}

    def config_output_slots():
        return {"hsv_image": DataType.ARRAY}

    def config_params():
        return {}  # No parameters needed for this transformation

    def process(self, rgb_image: Data):
        if rgb_image is None or rgb_image.data is None:
            return None

        # Extract RGB values
        r, g, b = rgb_image.data[..., 0], rgb_image.data[..., 1], rgb_image.data[..., 2]

        # Convert RGB to HSV
        hsv = np.vectorize(colorsys.rgb_to_hsv)(r, g, b)
        hsv_image = np.stack(hsv, axis=-1)

        return {"hsv_image": (hsv_image, {**rgb_image.meta})}



import pickle
import numpy as np
from goofi.data import Data, DataType
from goofi.node import Node
from goofi.params import IntParam, StringParam
from usearch.index import Index
import os


class VectorDB(Node):
    """
    A Goofi node for searching the top N labels from a vector embedding database.
    """

    def config_input_slots():
        return {
            "input_vector": DataType.ARRAY,
        }

    def config_output_slots():
        return {
            "top_labels": DataType.TABLE,
            "vectors": DataType.ARRAY,
        }

    def config_params():
        return {
            "Control": {
                "top_n": IntParam(10, 1, 100, doc="Number of top labels to return"),
                "database_path": StringParam("index_nouns.pkl", doc="Path to the database pickle file"),
            }
        }

    def setup(self):
        """
        Load the database index and mapping during initialization.
        """
        database_path = self.params.Control.database_path.value
        #database_path = os.path.join(self.assets_path, database_path)
        try:
            with open(database_path, "rb") as f:
                index_data, self.idx2word = pickle.load(f)
            self.index = Index.restore(index_data)
        except Exception as e:
            raise RuntimeError(f"Failed to load the database from {database_path}: {e}")

    def process(self, input_vector: Data):
        """
        Search for the top N labels in the vector embedding database.

        Args:
            input_vector (Data): Input vector to search for (1D array).

        Returns:
            dict: The top N labels with their distances and vectors.
        """
        if input_vector is None:
            return None

        input_vector_array = np.squeeze(input_vector.data)

        if input_vector_array.ndim != 1:
            raise ValueError("Input vector must be 1D.")

        # Ensure the input vector matches the index dimensionality
        index_dimensionality = self.index.ndim  # Assuming self.index has an 'ndim' attribute
        if input_vector_array.shape[0] != index_dimensionality:
            raise ValueError(
                f"Input vector dimensions ({input_vector_array.shape[0]}) do not match index dimensions ({index_dimensionality})."
            )

        top_n = self.params.Control.top_n.value

        # Perform the search
        search_results = self.index.search(input_vector_array, top_n, exact=True)

        # Retrieve vectors efficiently
        vectors = [self.index.get(result.key) for result in search_results]
        vectors = np.array(vectors)

        # Get the top N labels and their distances
        top_labels = {self.idx2word[result.key]: Data(DataType.ARRAY, result.distance, {}) for result in search_results}

        return {"top_labels": (top_labels, {}), "vectors": (vectors, {})}

    def Control_database_path_changed(self, value):
        self.setup()



from goofi.data import Data, DataType
from goofi.node import Node
from goofi.params import IntParam, StringParam


class SpectroMorphology(Node):
    def config_input_slots():
        return {"data": DataType.ARRAY}

    def config_output_slots():
        return {"spectro": DataType.ARRAY}

    def config_params():
        return {
            "Parameters": {
                "method": StringParam(
                    "SpectralCentroid",
                    options=[
                        "SpectralCentroid",
                        "SpectralCrestFactor",
                        "SpectralDecrease",
                        "SpectralFlatness",
                        "SpectralFlux",
                        "SpectralKurtosis",
                        "SpectralMfccs",
                        "SpectralPitchChroma",
                        "SpectralRolloff",
                        "SpectralSkewness",
                        "SpectralSlope",
                        "SpectralSpread",
                        "SpectralTonalPowerRatio",
                    ],
                ),
                "window": IntParam(100, 10, 10000),
                "overlap": IntParam(50, 1, 1000),
            }
        }

    def process(self, data: Data):
        """
        Convert a musical scale into a list of HSV colors based on the scale's frequency values
        and their averaged consonance.
        """
        if data is None:
            return None

        if data.data.ndim > 1:
            raise ValueError("Data must be 1D")

        method = self.params["Parameters"]["method"].value
        window = self.params["Parameters"]["window"].value
        overlap = self.params["Parameters"]["overlap"].value
        v, t = computeFeatureCl_new(data.data, method, data.meta["sfreq"], window=window, overlap=overlap)
        return {"spectro": (v, data.meta)}


pyACA_mod = None


def computeFeatureCl_new(afAudioData, cFeatureName, f_s, window=4000, overlap=1):
    """Calculate spectromorphological metrics on time series.

    Parameters
    ----------
    afAudioData : array (numDataPoints, )
        Input signal.
    cFeatureName : str
        {'SpectralCentroid', 'SpectralCrestFactor', 'SpectralDecrease',
         'SpectralFlatness', 'SpectralFlux', 'SpectralKurtosis',
         'SpectralMfccs', 'SpectralPitchChroma', 'SpectralRolloff',
         'SpectralSkewness', 'SpectralSlope', 'SpectralSpread',
         'SpectralTonalPowerRatio', 'TimeAcfCoeff', 'TimeMaxAcf',
         'TimePeakEnvelope', 'TimeRms', 'TimeStd', 'TimeZeroCrossingRate'}
    f_s : int
        Sampling frequency.
    window : int
        Length of the moving window in samples.
    overlap : int
        Overlap between each moving window in samples.

    Returns
    -------
    v : array
        Vector of the spectromorphological metric.
    t : array
        Timestamps.
    """
    global pyACA_mod
    if pyACA_mod is None:
        import pyACA

        pyACA_mod = pyACA

    [v, t] = pyACA_mod.computeFeature(cFeatureName, afAudioData, f_s, None, window, overlap)
    return (v, t)



import time
from typing import Any, Dict, Tuple

import numpy as np
import serial
import serial.tools.list_ports

from goofi.data import DataType
from goofi.node import Node
from goofi.params import IntParam


class SerialStream(Node):
    ESC = b"\xDB"
    END = b"\xC0"
    ESC_END = b"\xDC"
    ESC_ESC = b"\xDD"

    def config_params():
        return {"serial": {"sfreq": IntParam(512, 128, 1000), "port": ""}, "common": {"autotrigger": True}}

    def config_output_slots():
        return {"out": DataType.ARRAY}

    def setup(self):
        if self.params.serial.port.value:
            self.ser = serial.Serial(self.params.serial.port.value, 115200, timeout=1)
        else:
            port = self.detect_serial_port()
            if port is None:
                self.ser = None
                raise RuntimeError("No serial port found.")

            print(f"Found serial port {port}")
            self.ser = serial.Serial(port, 115200, timeout=1)

        self.last_time = None
        self.last_sample = None

    def process(self) -> Dict[str, Tuple[np.ndarray, Dict[str, Any]]]:
        data_buf, time_buf = [], []

        packet = bytearray()
        start = time.time()
        while time.time() - start < 1 / self.params.common.max_frequency.value:
            c = self.ser.read()
            if c == self.END:
                # ignore empty packets
                if packet and len(packet) == 2:
                    value = packet[0] << 8 | packet[1]
                    current_time = time.time()
                    data_buf.append(value)
                    time_buf.append(current_time)
                packet = bytearray()
            elif c == self.ESC:
                c = self.ser.read()
                if c == self.ESC_END:
                    packet.append(0xC0)
                elif c == self.ESC_ESC:
                    packet.append(0xDB)
                else:
                    print("Invalid escape sequence")
            elif len(c) > 0:
                packet.append(c[0])

        if len(data_buf) == 0:
            # no data received
            return None

        if self.last_time is None:
            # first time, just store the data
            self.last_time = time_buf[-1]
            self.last_sample = data_buf[-1]
            return None

        # resample the data
        dt = 1 / self.params.serial.sfreq.value
        xs = np.arange(time_buf[0], time_buf[-1], dt)
        # TODO: make sure the time array is correct and we don't have discontinuities
        data = np.interp(xs, time_buf, data_buf, left=self.last_sample)

        self.last_time = time_buf[-1]
        self.last_sample = data[-1]

        meta = {"sfreq": self.params.serial.sfreq.value}
        return {"out": (data, meta)}

    def detect_serial_port(self, names=["Arduino", "Serial"]):
        # detect the serial port
        ports = serial.tools.list_ports.comports()
        for port in ports:
            for name in names:
                if name in port.description:
                    return port.device
        return None

    def serial_port_changed(self, value: str) -> None:
        # reinitialize the client
        self.setup()



from math import gcd

import numpy as np

from goofi.data import Data, DataType
from goofi.node import Node
from goofi.params import IntParam


class Resample(Node):
    def config_input_slots():
        # Defining one input slot for one input signal
        return {"data": DataType.ARRAY}

    def config_output_slots():
        # Defining one output slot for the resampled signal
        return {"out": DataType.ARRAY}

    def config_params():
        return {
            "resample": {
                "new_sfreq": IntParam(1000, 10, 44100),  # New sampling frequency as a parameter
                "axis": IntParam(-1, -1, 4),  # Axis to resample the signal
            }
        }

    def setup(self):
        from scipy.signal import resample_poly

        self.resample_poly = resample_poly

    def process(self, data: Data):
        if data is None or data.data is None:
            print("Data is None")
            return None

        # Original sampling frequency from metadata
        sf = data.meta["sfreq"]

        # Retrieve new sampling frequency parameter
        new_sfreq = self.params["resample"]["new_sfreq"].value
        axis = self.params["resample"]["axis"].value

        # Calculate up and down factors based on the gcd of sf and new_sfreq
        factor = gcd(int(sf), int(new_sfreq))
        up = new_sfreq // factor
        down = sf // factor

        signal = np.array(data.data)

        # Check if the signal contains any NaN or inf values
        if np.any(np.isnan(signal)) or np.any(np.isinf(signal)):
            print("Signal contains NaN or inf values")

        # Resample the signal based on the specified axis
        resampled_signal = self.resample_poly(signal, up, down, axis=axis, padtype="line")

        # Adjust axis for negative values
        if axis < 0:
            axis += signal.ndim

        # Remove the corresponding channel metadata if it exists
        if f"dim{axis}" in data.meta["channels"]:
            del data.meta["channels"][f"dim{axis}"]

        # Update the 'sfreq' metadata
        data.meta["sfreq"] = new_sfreq

        return {"out": (resampled_signal, data.meta)}



import json
import math
import os
from os.path import exists, join

from goofi.data import Data, DataType
from goofi.node import Node
from goofi.params import StringParam


class EEGEmbedding(Node):
    def config_input_slots():
        return {"eeg": DataType.ARRAY}

    def config_output_slots():
        return {"embeddings": DataType.ARRAY}

    def config_params():
        return {
            "embedding": {
                "device": StringParam("auto", options=["auto", "cpu", "cuda"], doc="Device to run the model on"),
            }
        }

    def setup(self):
        import torch
        from safetensors.torch import load_file

        self.torch = torch

        model_path = join(self.assets_path, "eeg_encoder_55-95_40_classes")
        if not exists(model_path):
            import gdown

            model_url = "https://drive.google.com/drive/folders/1KAVbjwGdeA8RKTobMmzWNkHeKYklpCPF?usp=sharing"
            gdown.download_folder(model_url, output=model_path)

        # initialize the model classes
        EEGModelConfig, ChannelNetModel = init_model_classes()

        # load the model configuration and initialize the model
        config = EEGModelConfig.from_pretrained(model_path)
        self.model = ChannelNetModel(config)
        # load model weights from safetensors
        self.model.load_state_dict(load_file(join(model_path, "model.safetensors")))
        self.model.eval()

        # set the device to run the model on
        self.embedding_device_changed(self.params.embedding.device.value)

    def process(self, eeg: Data):
        if eeg is None:
            return None

        assert eeg.data.shape[0] == 128, f"Expected 128 channels, got {eeg.data.shape[0]}"
        assert eeg.data.shape[1] >= 440, f"Expected at least 440 samples, got {eeg.data.shape[1]}"
        assert eeg.meta["sfreq"] == 1000, f"Expected sampling frequency of 1000, got {eeg.meta['sfreq']}"

        # crop the EEG data to the last 440 samples
        eeg_data = eeg.data[:, -440:]
        # convert EEG data to PyTorch tensor and move to device
        eeg_data = self.torch.tensor(eeg_data, dtype=self.torch.float32)[None, None].to(self.device)

        # Extract embeddings from the EEG encoder
        with self.torch.inference_mode():
            # assert False, str(eeg_tensor.shape)
            emb_out = self.model(eeg_data)[0].squeeze()

        # Return the embeddings as output
        return {"embeddings": (emb_out.cpu().numpy(), {})}

    def embedding_device_changed(self, value):
        self.device = self.params.embedding.device.value
        if self.device == "auto":
            self.device = "cuda" if self.torch.cuda.is_available() else "cpu"

        self.model.to(self.device)


def init_model_classes():
    import torch
    from torch import nn
    from transformers import PretrainedConfig, PreTrainedModel

    # This is the model presented in the work: S. Palazzo, C. Spampinato, I. Kavasidis, D. Giordano, J. Schmidt, M. Shah, Decoding Brain Representations by
    # Multimodal Learning of Neural Activity and Visual Features,  IEEE TRANSACTIONS ON PATTERN ANALYSIS AND MACHINE INTELLIGENCE, 2020, doi: 10.1109/TPAMI.2020.2995909
    # code adapted from https://github.com/abhijitmishra/Thought2Text

    class EEGModelConfig(PretrainedConfig):
        model_type = "eeg_channelnet"

        def __init__(
            self,
            in_channels=1,
            temp_channels=10,
            out_channels=50,
            num_classes=40,
            embedding_size=512,
            input_width=440,
            input_height=128,
            temporal_dilation_list=None,
            temporal_kernel=(1, 33),
            temporal_stride=(1, 2),
            num_temp_layers=4,
            num_spatial_layers=4,
            spatial_stride=(2, 1),
            num_residual_blocks=4,
            down_kernel=3,
            down_stride=2,
            **kwargs,
        ):
            if temporal_dilation_list is None:
                temporal_dilation_list = [(1, 1), (1, 2), (1, 4), (1, 8), (1, 16)]

            super().__init__(**kwargs)
            self.in_channels = in_channels
            self.temp_channels = temp_channels
            self.out_channels = out_channels
            self.num_classes = num_classes
            self.embedding_size = embedding_size
            self.input_width = input_width
            self.input_height = input_height
            self.temporal_dilation_list = temporal_dilation_list
            self.temporal_kernel = temporal_kernel
            self.temporal_stride = temporal_stride
            self.num_temp_layers = num_temp_layers
            self.num_spatial_layers = num_spatial_layers
            self.spatial_stride = spatial_stride
            self.num_residual_blocks = num_residual_blocks
            self.down_kernel = down_kernel
            self.down_stride = down_stride

        @classmethod
        def from_dict(cls, config_dict, **kwargs):
            """Creates a config from a dictionary."""
            return cls(**config_dict, **kwargs)

        @classmethod
        def from_pretrained(cls, model_path, **kwargs):
            """Loads the configuration from a file."""
            config_file = f"{model_path}/config.json"
            with open(config_file, "r") as f:
                config_dict = json.load(f)
            return cls.from_dict(config_dict, **kwargs)

    class FeaturesExtractor(nn.Module):
        def __init__(self, config):
            super().__init__()

            self.temporal_block = TemporalBlock(
                config.in_channels,
                config.temp_channels,
                config.num_temp_layers,
                config.temporal_kernel,
                config.temporal_stride,
                config.temporal_dilation_list,
                config.input_width,
            )

            self.spatial_block = SpatialBlock(
                config.temp_channels * config.num_temp_layers,
                config.out_channels,
                config.num_spatial_layers,
                config.spatial_stride,
                config.input_height,
            )

            self.res_blocks = nn.ModuleList(
                [
                    nn.Sequential(
                        ResidualBlock(
                            config.out_channels * config.num_spatial_layers,
                            config.out_channels * config.num_spatial_layers,
                        ),
                        ConvLayer2D(
                            config.out_channels * config.num_spatial_layers,
                            config.out_channels * config.num_spatial_layers,
                            config.down_kernel,
                            config.down_stride,
                            0,
                            1,
                        ),
                    )
                    for i in range(config.num_residual_blocks)
                ]
            )

            self.final_conv = ConvLayer2D(
                config.out_channels * config.num_spatial_layers,
                config.out_channels,
                config.down_kernel,
                1,
                0,
                1,
            )

        def forward(self, x):
            out = self.temporal_block(x)

            out = self.spatial_block(out)

            if len(self.res_blocks) > 0:
                for res_block in self.res_blocks:
                    out = res_block(out)

            out = self.final_conv(out)

            return out

    class ChannelNetModel(PreTrainedModel):
        """The model for EEG classification.
        The imput is a tensor where each row is a channel the recorded signal and each colums is a time sample.
        The model performs different 2D to extract temporal e spatial information.
        The output is a vector of classes where the maximum value is the predicted class.
        Args:
            in_channels: number of input channels
            temp_channels: number of features of temporal block
            out_channels: number of features before classification
            num_classes: number possible classes
            embedding_size: size of the embedding vector
            input_width: width of the input tensor (necessary to compute classifier input size)
            input_height: height of the input tensor (necessary to compute classifier input size)
            temporal_dilation_list: list of dilations for temporal convolutions, second term must be even
            temporal_kernel: size of the temporal kernel, second term must be even (default: (1, 32))
            temporal_stride: size of the temporal stride, control temporal output size (default: (1, 2))
            num_temp_layers: number of temporal block layers
            num_spatial_layers: number of spatial layers
            spatial_stride: size of the spatial stride
            num_residual_blocks: the number of residual blocks
            down_kernel: size of the bottleneck kernel
            down_stride: size of the bottleneck stride
        """

        def __init__(self, config: EEGModelConfig):
            super().__init__(config=config)

            self.encoder = FeaturesExtractor(config=config)

            encoding_size = (
                self.encoder(torch.zeros(1, config.in_channels, config.input_height, config.input_width))
                .contiguous()
                .view(-1)
                .size()[0]
            )
            self.projector = nn.Linear(encoding_size, config.embedding_size)
            self.classifier = nn.Linear(config.embedding_size, config.num_classes)

        def forward(self, x):
            out = self.encoder(x)

            out = out.view(x.size(0), -1)
            emb = self.projector(out)

            cls = self.classifier(emb)

            return emb, cls

    class ConvLayer2D(nn.Sequential):
        def __init__(self, in_channels, out_channels, kernel, stride, padding, dilation):
            super().__init__()
            self.add_module("norm", nn.BatchNorm2d(in_channels))
            self.add_module("relu", nn.ReLU(True))
            self.add_module(
                "conv",
                nn.Conv2d(
                    in_channels, out_channels, kernel_size=kernel, stride=stride, padding=padding, dilation=dilation, bias=True
                ),
            )
            self.add_module("drop", nn.Dropout2d(0.2))

        def forward(self, x):
            return super().forward(x)

    class TemporalBlock(nn.Module):
        def __init__(self, in_channels, out_channels, n_layers, kernel_size, stride, dilation_list, in_size):
            super().__init__()
            if len(dilation_list) < n_layers:
                dilation_list = dilation_list + [dilation_list[-1]] * (n_layers - len(dilation_list))

            padding = []
            # Compute padding for each temporal layer to have a fixed size output
            # Output size is controlled by striding to be 1 / 'striding' of the original size
            for dilation in dilation_list:
                filter_size = kernel_size[1] * dilation[1] - 1
                temp_pad = math.floor((filter_size - 1) / 2) - 1 * (dilation[1] // 2 - 1)
                padding.append((0, temp_pad))

            self.layers = nn.ModuleList(
                [
                    ConvLayer2D(in_channels, out_channels, kernel_size, stride, padding[i], dilation_list[i])
                    for i in range(n_layers)
                ]
            )

        def forward(self, x):
            features = []

            for layer in self.layers:
                out = layer(x)
                features.append(out)

            out = torch.cat(features, 1)
            return out

    class SpatialBlock(nn.Module):
        def __init__(self, in_channels, out_channels, num_spatial_layers, stride, input_height):
            super().__init__()

            kernel_list = []
            for i in range(num_spatial_layers):
                kernel_list.append(((input_height // (i + 1)), 1))

            padding = []
            for kernel in kernel_list:
                temp_pad = math.floor((kernel[0] - 1) / 2)  # - 1 * (kernel[1] // 2 - 1)
                padding.append((temp_pad, 0))

            feature_height = input_height // stride[0]

            self.layers = nn.ModuleList(
                [
                    ConvLayer2D(in_channels, out_channels, kernel_list[i], stride, padding[i], 1)
                    for i in range(num_spatial_layers)
                ]
            )

        def forward(self, x):
            features = []

            for layer in self.layers:
                out = layer(x)
                features.append(out)

            out = torch.cat(features, 1)

            return out

    def conv3x3(in_channels, out_channels, stride=1):
        return nn.Conv2d(in_channels, out_channels, kernel_size=3, stride=stride, padding=1, bias=False)

    # Residual block
    class ResidualBlock(nn.Module):
        def __init__(self, in_channels, out_channels, stride=1, downsample=None):
            super(ResidualBlock, self).__init__()
            self.conv1 = conv3x3(in_channels, out_channels, stride)
            self.bn1 = nn.BatchNorm2d(out_channels)
            self.relu = nn.ReLU(inplace=True)
            self.conv2 = conv3x3(out_channels, out_channels)
            self.bn2 = nn.BatchNorm2d(out_channels)
            self.downsample = downsample

        def forward(self, x):
            residual = x
            out = self.conv1(x)
            out = self.bn1(out)
            out = self.relu(out)
            out = self.conv2(out)
            out = self.bn2(out)
            if self.downsample:
                residual = self.downsample(x)
            out += residual
            out = self.relu(out)
            return out

    return EEGModelConfig, ChannelNetModel



import numpy as np

from goofi.data import Data, DataType
from goofi.node import Node
from goofi.params import FloatParam, IntParam, StringParam


class Fractality(Node):
    def config_input_slots():
        return {"data_input": DataType.ARRAY}

    def config_output_slots():
        return {"fractal_dimension": DataType.ARRAY}

    def config_params():
        return {
            "method": {
                "name": StringParam(
                    "hurst",
                    options=[
                        "hurst",
                        "fractal_correlation",
                        "fractal_petrosian",
                        "fractal_psdslope",
                        "fractal_higuchi",
                        "fractal_katz",
                        "fractal_linelength",
                        "fractal_nld",
                        "box_counting_2d",
                    ],
                ),
            },
            "box_counting": {
                "scales_base": FloatParam(2.0, 1.1, 10.0, doc="Base for the logspace calculation in box_counting"),
                "scales_start": FloatParam(0.01, 0.001, 1.0, doc="Start of the logspace in box_counting"),
                "scales_num": IntParam(10, 5, 100, doc="Number of steps in the logspace for box_counting"),
            },
            "fractal_higuchi": {"k_max": IntParam(10, 1, 100, doc="Maximum k value for fractal_higuchi method")},
            "fractal_correlation": {
                "delay": IntParam(1, 1, 100, doc="Delay for fractal_correlation method"),
                "dimension": IntParam(2, 1, 100, doc="Dimension for fractal_correlation method"),
            },
        }

    def setup(self):
        import neurokit2 as nk

        self.nk = nk

    def process(self, data_input: Data):
        if data_input is None or data_input.data is None:
            return None

        method = self.params["method"]["name"].value

        # For methods in neurokit2
        if method == "fractal_katz":
            result, _ = self.nk.fractal_katz(data_input.data)
        elif method == "fractal_petrosian":
            result, _ = self.nk.fractal_petrosian(data_input.data)
        elif method == "fractal_linelength":
            result, _ = self.nk.fractal_linelength(data_input.data)
        elif method == "fractal_psdslope":
            result, _ = self.nk.fractal_psdslope(data_input.data)
        elif method == "fractal_nld":
            result, _ = self.nk.fractal_nld(data_input.data)
        elif method == "fractal_higuchi":
            result, _ = self.nk.fractal_higuchi(data_input.data, k_max=self.params["fractal_higuchi"]["k_max"].value)
        elif method == "hurst":
            result, _ = self.nk.fractal_hurst(data_input.data)
        elif method == "fractal_correlation":
            result, _ = self.nk.fractal_correlation(data_input.data)
        elif method == "box_counting_2d":
            result = self.box_counting(data_input.data)
            print(result)

        return {"fractal_dimension": (result, {})}

    def box_counting(self, data):
        # Check if data is either 2D or 3D with shape (x, y, 3)
        if len(data.shape) not in [2, 3] or (len(data.shape) == 3 and data.shape[2] != 3):
            raise ValueError("Data for box_counting should be either 2D or have a shape of (x, y, 3)")

        # Convert from RGB to grayscale if data is 3D
        if len(data.shape) == 3:
            data = self.rgb2gray(data)
        # Threshold the data to create a binary representation
        threshold = np.mean(data)
        data = (data > threshold).astype(int)

        if len(data.shape) != 2:
            raise ValueError("Data for box_counting should be 2D after any necessary conversions")

        # Find all the non-zero pixels (modify this if your image isn't binary)
        pixels = np.column_stack(np.where(data > 0))

        # Retrieve the scales parameters
        scales_base = self.params["box_counting"]["scales_base"].value
        scales_start = self.params["box_counting"]["scales_start"].value
        scales_num = self.params["box_counting"]["scales_num"].value

        scales = np.logspace(scales_start, 1, num=scales_num, endpoint=False, base=scales_base)
        Ns = []

        for scale in scales:
            H, _ = np.histogramdd(pixels, bins=(np.arange(0, data.shape[0], scale), np.arange(0, data.shape[1], scale)))
            Ns.append(np.sum(H > 0))

        coeffs = np.polyfit(np.log(scales), np.log(Ns), 1)
        return -coeffs[0]

    def rgb2gray(self, rgb):
        gray = 0.2989 * rgb[:, :, 0] + 0.5870 * rgb[:, :, 1] + 0.1140 * rgb[:, :, 2]
        return gray / 255  # Normalize to [0,1]


# https://github.com/ChatzigeorgiouGroup/FractalDimension/blob/master/FractalDimension.py



import time

import numpy as np

from goofi.data import Data, DataType
from goofi.node import Node
from goofi.params import BoolParam, IntParam, StringParam


class StaticBaseline(Node):
    def config_input_slots():
        return {"data": DataType.ARRAY}

    def config_output_slots():
        return {"normalized": DataType.ARRAY}

    def config_params():
        return {
            "baseline": {
                "n_seconds": IntParam(30, 1, 120),
                "method": StringParam("quantile", options=["quantile", "mean"]),
                "baseline_computation": BoolParam(trigger=True),
            }
        }

    def setup(self):
        from scipy.stats import rankdata

        self.rankdata = rankdata

        self.window = []
        self.time_origin = None

    def process(self, data: Data):
        if data is None or data.data is None:
            return None

        val = np.asarray(data.data)
        if val.ndim > 2:
            print("Error: StaticBaseline only accepts 1D or 2D arrays")
            return None

        normalized_value = np.zeros_like(val)

        # If baseline computation is triggered, reset the window and set time_origin
        if self.params["baseline"]["baseline_computation"].value:
            self.window = []
            self.time_origin = time.time()

        if self.time_origin:
            elapsed_time = time.time() - self.time_origin

            if elapsed_time < self.params["baseline"]["n_seconds"].value:
                self.window.extend(val)
            else:
                self.time_origin = None  # Reset time_origin after accumulating for n_seconds

        # Only compute the Z-score when we have enough data in our window
        if len(self.window) >= self.params["baseline"]["n_seconds"].value:
            # Compute the Z-score using the desired method
            if self.params["baseline"]["method"].value == "quantile":
                normalized_value = self.quantile_transform(val)
            elif self.params["baseline"]["method"].value == "mean":
                normalized_value = self.zscore(val)
        else:
            normalized_value = val  # You can choose to normalize or just use raw data when the window is not full

        return {"normalized": (normalized_value, data.meta)}

    def zscore(self, val):
        if val.ndim == 1:
            mean = np.mean(self.window)
            std = np.std(self.window)
            return (val - mean) / (std + 1e-8)
        elif val.ndim == 2:
            normalized = np.zeros_like(val)
            for i in range(val.shape[0]):
                mean = np.mean(self.window[i])
                std = np.std(self.window[i])
                normalized[i, :] = (val[i, :] - mean) / (std + 1e-8)
            return normalized

    def quantile_transform(self, val):
        # Check for dimension
        if val.ndim == 2:
            normalized = np.zeros_like(val)
            for i in range(val.shape[0]):
                normalized[i, :] = self._quantile_transform_1D(val[i, :])
            return normalized
        else:
            return self._quantile_transform_1D(val)

    def _quantile_transform_1D(self, arr):
        # Convert data to ranks
        ranks = self.rankdata(arr)
        # Scale ranks to [0, 1]
        scaled_ranks = (ranks - 1) / (len(arr) - 1)
        # Calculate quantile values
        quantiles = np.percentile(arr, 100 * scaled_ranks)
        # Return normalized quantile values for the input data
        return (arr - np.mean(quantiles)) / (np.std(quantiles) + 1e-8)



from goofi.data import Data, DataType
from goofi.node import Node
from goofi.params import StringParam


class Switch(Node):
    def config_input_slots():
        return {
            "selector": DataType.ARRAY,  # Input 1: Selector array
            "array1": DataType.ARRAY,  # Input 2: First array input
            "array2": DataType.ARRAY,  # Input 3: Second array input
            "array3": DataType.ARRAY,  # Input 4: Third array input
            "string1": DataType.STRING,  # Input 5: First string input
            "string2": DataType.STRING,  # Input 6: Second string input
            "string3": DataType.STRING,  # Input 7: Third string input
        }

    def config_output_slots():
        return {
            "array_out": DataType.ARRAY,  # Output 1: Array output
            "string_out": DataType.STRING,  # Output 2: String output
        }

    def config_params():
        return {
            "settings": {
                "mode": StringParam("array", ["array", "string"]),  # Dropdown to choose mode
            }
        }

    def process(self, selector: Data, array1: Data, array2: Data, array3: Data, string1: Data, string2: Data, string3: Data):
        if selector is None or selector.data is None:
            return

        selector_value = int(selector.data[0])  # Assume selector is a single-element array
        mode = self.params["settings"]["mode"].value

        if mode == "array":
            if selector_value == 1:
                if array1 is not None:
                    return {"array_out": (array1.data, {}), "string_out": None}
                return
            elif selector_value == 2:
                if array2 is not None:
                    return {"array_out": (array2.data, {}), "string_out": None}
                return
            elif selector_value == 3:
                if array3 is not None:
                    return {"array_out": (array3.data, {}), "string_out": None}
                return
            else:
                raise ValueError("Selector value must be 1 or 2")

        elif mode == "string":
            if selector_value == 1:
                if string1 is not None:
                    return {"array_out": None, "string_out": (string1.data, {})}
                return
            elif selector_value == 2:
                if string2 is not None:
                    return {"array_out": None, "string_out": (string2.data, {})}
                return
            elif selector_value == 3:
                if string3 is not None:
                    return {"array_out": None, "string_out": (string3.data, {})}
                return
            else:
                raise ValueError("Selector value must be 1 or 2")



import numpy as np

from goofi.data import Data, DataType
from goofi.node import Node
from goofi.params import BoolParam, FloatParam, IntParam, StringParam


class DimensionalityReduction(Node):
    def config_input_slots():
        return {"data": DataType.ARRAY, "new_data": DataType.ARRAY}

    def config_output_slots():
        return {
            "transformed": DataType.ARRAY,
            "new_components": DataType.ARRAY,
        }

    def config_params():
        return {
            "Control": {
                "reset": BoolParam(False, trigger=True, doc="Reset the buffer"),
                "method": StringParam("PCA", options=["PCA", "t-SNE", "UMAP"], doc="Dimensionality reduction method"),
                "n_components": IntParam(2, 1, 10, doc="Number of output dimensions"),
                "tsne_perplexity": FloatParam(30.0, 5.0, 50.0, doc="t-SNE perplexity"),
                "umap_neighbors": FloatParam(15, 2, 100, doc="Number of UMAP neighbors"),
            }
        }

    def setup(self):
        from sklearn.decomposition import PCA
        from sklearn.manifold import TSNE
        from umap import UMAP

        self.tsne_cls = TSNE
        self.pca_cls = PCA
        self.umap_cls = UMAP

        self.model = None
        self.components = None
        self.meta = None

    def process(self, data: Data, new_data: Data):
        if data is None:
            return None

        method = self.params.Control.method.value
        data_array = np.squeeze(data.data)

        if self.params.Control.reset.value:
            self.model = None
            self.components = None
            self.meta = None

        if self.components is not None:
            new_components = None
            if new_data is not None and self.model is not None:
                if method == "t-SNE":
                    raise ValueError("The t-SNE algorithm does not support transforming new data")

                new_components = self.model.transform(new_data.data)
                new_meta = new_data.meta
                if "channels" in new_meta and "dim1" in new_meta["channels"]:
                    del new_meta["channels"]["dim1"]

            return {
                "transformed": (self.components, self.meta),
                "new_components": (new_components, new_meta) if new_components is not None else None,
            }

        if data_array.ndim != 2:
            raise ValueError("Data must be 2D")

        n_components = int(self.params.Control.n_components.value)
        perplexity = self.params.Control.tsne_perplexity.value
        n_neighbors = int(self.params.Control.umap_neighbors.value)

        self.meta = data.meta
        if "channels" in self.meta and "dim1" in self.meta["channels"]:
            del self.meta["channels"]["dim1"]

        new_components = None
        if method == "PCA":
            self.model = self.pca_cls(n_components=n_components)
            self.components = self.model.fit_transform(data_array)

            if new_data is not None:
                new_components = self.model.transform(new_data.data)
                new_meta = new_data.meta
                if "channels" in new_meta and "dim1" in new_meta["channels"]:
                    del new_meta["channels"]["dim1"]

        elif method == "t-SNE":
            self.model = self.tsne_cls(n_components=n_components, perplexity=perplexity, init="pca", random_state=42)
            self.components = self.model.fit_transform(data_array)

        elif method == "UMAP":
            # Initialize UMAP model
            self.model = self.umap_cls(n_neighbors=n_neighbors, n_components=n_components, random_state=42)
            self.components = self.model.fit_transform(data_array)

            if new_data is not None:
                new_components = self.model.transform(new_data.data)
                new_meta = new_data.meta
                if "channels" in new_meta and "dim1" in new_meta["channels"]:
                    del new_meta["channels"]["dim1"]

        return {
            "transformed": (self.components, self.meta),
            "new_components": (new_components, self.meta) if new_components is not None else None,
        }



import logging

import cv2
import numpy as np

from goofi.data import Data, DataType
from goofi.node import Node
from goofi.params import BoolParam, IntParam, StringParam


class FacialExpression(Node):
    def config_input_slots():
        return {"image": DataType.ARRAY}

    def config_output_slots():
        return {"emotion_probabilities": DataType.ARRAY, "action_units": DataType.ARRAY, "main_emotion": DataType.STRING}

    def config_params():
        return {
            "emotion_recognition": {
                "device": StringParam(
                    "auto", options=["cpu", "cuda", "mps", "auto"], doc="Device to use for emotion recognition"
                ),
                "reduction_factor": IntParam(1, 1, 10, doc="Factor to reduce the image size by before processing"),
                "disable_neutral": BoolParam(False, doc="Whether to disable neutral emotion"),
                "threshold": IntParam(0, 0, 100, doc="Threshold for emotion recognition"),
            }
        }

    def setup(self):
        try:
            from feat import Detector
        except ImportError:
            raise ImportError(
                "The dependency 'py-feat', which requires pytorch, is not installed. "
                "Please install it using 'pip install py-feat', and make sure the pytorch version is "
                "compatible with your system."
            )

        logging.basicConfig(level=logging.INFO)
        # Load the py-feat detector with the specified models
        self.detector = Detector(
            face_model="retinaface",
            landmark_model="mobilefacenet",
            au_model="xgb",
            emotion_model="resmasknet",
            facepose_model="img2pose",
            device=self.params["emotion_recognition"]["device"].value,
            n_jobs=1,
        )
        self.emotion_names = ["anger", "disgust", "fear", "happiness", "sadness", "surprise", "neutral"]
        self.last_emotion = None

    def process(self, image: Data):
        if image is None or image.data is None:
            logging.warning("No image data to process.")
            return {
                "emotion_probabilities": None,
                "face_box": None,
                "landmarks": None,
                "action_units": None,
                "main_emotion": None,
            }

        # Convert image data to the format expected by py-feat Detector
        if image.data.dtype != np.uint8:
            image_data = ((image.data - image.data.min()) * (255.0 / (image.data.max() - image.data.min()))).astype(np.uint8)
        else:
            image_data = image.data
        image_data = cv2.cvtColor(image_data, cv2.COLOR_BGR2RGB)

        # rescale image data by a factor
        factor = self.params["emotion_recognition"]["reduction_factor"].value
        image_data = cv2.resize(image_data, (image_data.shape[1] // factor, image_data.shape[0] // factor))
        # Detect faces and their attributes using the image data
        detected_faces = self.detector.detect_faces(image_data)
        if not detected_faces:
            logging.warning("No face detected in the image.")
            return {"emotion_probabilities": None, "action_units": None, "main_emotion": None}

        detected_landmarks = self.detector.detect_landmarks(image_data, detected_faces)

        # Detect action units
        if detected_landmarks:
            action_units = self.detector.detect_aus(image_data, detected_landmarks)
        else:
            action_units = None

        # Detect emotions
        emotions = self.detector.detect_emotions(image_data, detected_faces, detected_landmarks)
        threshold = self.params["emotion_recognition"]["threshold"].value
        # Process the emotions array
        if emotions is not None and len(emotions) > 0:
            if self.params["emotion_recognition"]["disable_neutral"].value:
                emotions[0][0][6] = 0
            emotion_probabilities = emotions[0][0]  # Assuming you're interested in the first detected face
            main_emotion_index = np.argmax(emotion_probabilities)
            main_emotion = self.emotion_names[main_emotion_index]
            if emotion_probabilities[main_emotion_index] < threshold / 100:
                if self.last_emotion is not None:
                    main_emotion = self.last_emotion
                else:
                    main_emotion = ""
            self.last_emotion = main_emotion
        else:
            emotion_probabilities = np.array([0, 0, 0, 0, 0, 0, 0])
            main_emotion = ""

        return {
            "emotion_probabilities": (np.array(emotion_probabilities), {}),
            "action_units": (np.array(action_units[0]), {}),
            "main_emotion": (main_emotion, {}),
        }



import numpy as np
from goofi.data import Data, DataType
from goofi.node import Node
from goofi.params import FloatParam, BoolParam

class Walker(Node):
    """
    A Goofi node that simulates walking on the globe based on initial coordinates,
    an input angle, and velocity. Outputs the new coordinates, handling crossing
    over the poles and the International Date Line.
    """

    def config_input_slots():
        return {
            "angle": DataType.ARRAY,
            "velocity": DataType.ARRAY,
            "water": DataType.ARRAY
        }

    def config_output_slots():
        return {
            "latitude": DataType.ARRAY,
            "longitude": DataType.ARRAY
        }

    def config_params():
        return {
            "initial_coordinates": {
                "latitude": FloatParam(0.0, -90, 90),
                "longitude": FloatParam(0.0, -180, 180),
                "reset": BoolParam(False, trigger=True),
                "water_speed_factor": FloatParam(5.0, 0.1, 10.0)
                
            },
        }

    def setup(self):
        self.lat, self.lon = self.params["initial_coordinates"]["latitude"].value, \
                   self.params["initial_coordinates"]["longitude"].value
    def process(self, angle: Data, velocity: Data, water: Data):
        """
        Processes the input angle and velocity to compute new geographical coordinates,
        considering the curvature of the Earth and boundaries such as the poles and the
        International Date Line.
        """
        if angle is None or velocity is None:
            return None

        if self.params["initial_coordinates"]["reset"].value:
            self.lat, self.lon = self.params["initial_coordinates"]["latitude"].value, \
                                 self.params["initial_coordinates"]["longitude"].value

        if water is None:
            water_situation = 0
        else:
            water_situation = water.data
        
        water_speed_factor = self.params["initial_coordinates"]["water_speed_factor"].value
        if water_situation == 1:
            velocity.data = velocity.data * water_speed_factor
        # Convert angle to radians for computation
        angle_rad = np.radians(angle.data)

        # Calculate new latitude and longitude
        self.lat = self.lat + velocity.data * np.cos(angle_rad)
        self.lon = self.lon + velocity.data * np.sin(angle_rad)

        # Handle crossing the North/South poles
        if self.lat > 90:
            self.lat = 180 - self.lat
            #self.lon += 180
        elif self.lat < -90:
            self.lat = -180 - self.lat
            #self.lon += 180

        # Normalize longitude to be within -180 to 180
        if self.lon > 180:
            self.lon -= 360
        elif self.lon < -180:
            self.lon += 360

        return {"latitude": (self.lat, {}), "longitude": (self.lon, {})}




import numpy as np
from goofi.data import Data, DataType
from goofi.node import Node
from goofi.params import IntParam, StringParam


class TuningReduction(Node):
    def config_input_slots():
        return {"tuning": DataType.ARRAY}

    def config_output_slots():
        return {
            "reduced": DataType.ARRAY,
        }

    def config_params():
        return {
            "Mode_Generation": {
                "n_steps": IntParam(5, 2, 20, doc="Number of steps in the musical mode output"),
                "function": StringParam("harmsim", options=["harmsim", "cons", "denom"], doc="Harmonicity methods"),
            }
        }

    def setup(self):
        from biotuner.scale_construction import create_mode
        from biotuner.metrics import dyad_similarity, metric_denom, compute_consonance

        self.create_mode = create_mode
        self.dyad_similarity = dyad_similarity
        self.metric_denom = metric_denom
        self.compute_consonance = compute_consonance

    def process(self, tuning: Data):
        if tuning is None:
            return None

        tuning.data = np.squeeze(tuning.data)
        if tuning.data.ndim > 1:
            raise ValueError("Data must be 1D")

        n_steps = self.params["Mode_Generation"]["n_steps"].value
        function = self.params["Mode_Generation"]["function"].value
        if function == "harmsim":
            reduced = self.create_mode(tuning.data, n_steps, self.dyad_similarity)
        if function == "denom":
            reduced = self.create_mode(tuning.data, n_steps, self.metric_denom)
        if function == "cons":
            reduced = self.create_mode(tuning.data, n_steps, self.compute_consonance)
        return {
            "reduced": (np.array(reduced), tuning.meta),
        }



import soundfile as sf

from goofi.data import Data, DataType
from goofi.node import InputSlot, Node
from goofi.params import FloatParam, IntParam, StringParam


class Audio2Txt(Node):
    @staticmethod
    def config_input_slots():
        return {
            "prompt": InputSlot(DataType.STRING, trigger_process=False),
            "audio": DataType.ARRAY,
        }

    @staticmethod
    def config_output_slots():
        return {"generated_text": DataType.STRING}

    @staticmethod
    def config_params():
        return {
            "audio_to_text": {
                "provider": StringParam("huggingface", options=["huggingface", "nexa"], doc="Provider for audio LMs"),
                "model": StringParam(
                    "Qwen/Qwen2-Audio-7B-Instruct",
                    doc="Huggingface model ID for audio captioning",
                ),
                "max_new_tokens": IntParam(30, 10, 1024, doc="Maximum number of tokens to generate"),
                "temperature": FloatParam(0.7, 0.1, 2.0, doc="Sampling temperature for text generation"),
            }
        }

    def setup(self):
        import librosa

        self.librosa = librosa

        provider = self.params.audio_to_text.provider.value
        if provider == "huggingface":
            self.setup_huggingface()
        elif provider == "nexa":
            self.setup_nexa()

    def process(self, prompt: Data, audio: Data):
        if audio.data is None:
            return None

        prompt = prompt.data

        if self.params.audio_to_text.provider.value == "huggingface":
            generated_text = self.generate_huggingface(prompt, audio)
        elif self.params.audio_to_text.provider.value == "nexa":
            generated_text = self.generate_nexa(prompt, audio)
        else:
            raise ValueError(f"Unsupported provider: {self.params.audio_to_text.provider.value}")

        return {"generated_text": (generated_text, {})}

    def audio_to_text_provider_changed(self, value):
        self.setup()

    def audio_to_text_model_changed(self, value):
        self.setup()

    def setup_huggingface(self):
        import torch

        try:
            from transformers import AutoProcessor, Qwen2AudioForConditionalGeneration
        except ModuleNotFoundError:
            raise ModuleNotFoundError("Please install transformers to use Qwen2-Audio models via pip install transformers")

        try:
            self.processor = AutoProcessor.from_pretrained(
                self.params["audio_to_text"]["model"].value, torch_dtype=torch.float16
            )
            self.model = Qwen2AudioForConditionalGeneration.from_pretrained(
                self.params["audio_to_text"]["model"].value,
                device_map="cuda",
                torch_dtype=torch.float16,
            )
        except Exception as e:
            print(f"Error initializing Huggingface model: {e}")
            raise

    def generate_huggingface(self, prompt, audio):
        # Ensure the audio is sampled at the model's expected rate
        sampling_rate = self.processor.feature_extractor.sampling_rate
        audio_array = self.librosa.resample(audio.data, orig_sr=audio.meta["sfreq"], target_sr=sampling_rate)

        # Prepare the conversation input
        conversation = [
            {
                "role": "user",
                "content": [
                    {"type": "audio", "audio_url": "embedded"},
                    {"type": "text", "text": prompt},
                ],
            },
        ]
        text = self.processor.apply_chat_template(conversation, add_generation_prompt=True, tokenize=False)

        inputs = self.processor(
            text=text,
            audios=[audio_array],
            sampling_rate=sampling_rate,
            return_tensors="pt",
            padding=True,
        )
        inputs = {k: v.to(self.model.device) for k, v in inputs.items()}

        # Generate text
        generate_ids = self.model.generate(
            **inputs,
            max_new_tokens=self.params.audio_to_text.max_new_tokens.value,
            temperature=self.params.audio_to_text.temperature.value,
        )
        generate_ids = generate_ids[:, inputs["input_ids"].size(1) :]

        generated_text = self.processor.batch_decode(
            generate_ids, skip_special_tokens=True, clean_up_tokenization_spaces=False
        )[0]
        return generated_text

    def setup_nexa(self):
        try:
            from nexa.gguf.nexa_inference_audio_lm import NexaAudioLMInference
        except Exception as e:
            print(f"Error initializing Nexa model: {e}")
            raise

        self.model = NexaAudioLMInference(self.params.audio_to_text.model.value)

    def generate_nexa(self, prompt, audio):
        # save audio to temp file (nexa expects 16kHz audio)
        audio_file = "./temp_audio.wav"
        sf.write(audio_file, self.librosa.resample(audio.data, orig_sr=audio.meta["sfreq"], target_sr=16000), 16000)

        # TODO: this doesn't work due to current limitiation in Nexa
        self.model.params["max_new_tokens"] = self.params.audio_to_text.max_new_tokens.value
        self.model.params["temperature"] = self.params.audio_to_text.temperature.value

        # generate text
        response = self.model.inference(audio_file, prompt)

        # TODO: avoid reloading the model for each inference, the model should stay loaded in memory but currently it seems to reload every time so we need to cleanup
        self.model.cleanup()

        return response



import numpy as np
from goofi.node import Node
from goofi.data import DataType, Data
import colorsys


class HSVtoRGB(Node):
    def config_input_slots():
        return {"hsv_image": DataType.ARRAY}

    def config_output_slots():
        return {"rgb_image": DataType.ARRAY}

    def config_params():
        return {}  # No parameters needed for this transformation

    def process(self, hsv_image: Data):
        if hsv_image is None or hsv_image.data is None:
            return None

        # Extract HSV values
        h, s, v = hsv_image.data[..., 0], hsv_image.data[..., 1], hsv_image.data[..., 2]

        # Convert HSV to RGB
        rgb = np.vectorize(colorsys.hsv_to_rgb)(h, s, v)
        rgb_image = np.stack(rgb, axis=-1)

        return {"rgb_image": (rgb_image, {**hsv_image.meta})}



import numpy as np

from goofi.data import Data, DataType
from goofi.node import Node
from goofi.params import FloatParam


class Clip(Node):
    def config_input_slots():
        return {"array": DataType.ARRAY}

    def config_output_slots():
        return {"out": DataType.ARRAY}

    def config_params():
        return {"clip": {"min": FloatParam(-1), "max": FloatParam(1)}}

    def process(self, array: Data):
        if array is None:
            return None

        result = np.clip(array.data, self.params.clip.min.value, self.params.clip.max.value)

        return {"out": (result, array.meta)}



import numpy as np

from goofi.data import Data, DataType
from goofi.node import Node
from goofi.params import StringParam


class CardioRespiratoryVariability(Node):
    def config_input_slots():
        return {"data": DataType.ARRAY}

    def config_output_slots():
        return {
            "Mean": DataType.ARRAY,
            "SDNN": DataType.ARRAY,
            "SDSD": DataType.ARRAY,
            "RMSSD": DataType.ARRAY,
            "VLF": DataType.ARRAY,
            "LF": DataType.ARRAY,
            "HF": DataType.ARRAY,
            "LF/HF": DataType.ARRAY,
            #"LZC": DataType.ARRAY,
            "Peaks": DataType.ARRAY,
            "Rate": DataType.ARRAY
        }

    def config_params():
        return {
            "cardiorespiratory": {
                "input_type": StringParam("ppg", options=["ppg", "ecg", "rsp"]),
            },
        }

    def setup(self):
        import neurokit2 as nk

        self.neurokit = nk

    def process(self, data: Data):
        if data is None or data.data is None:
            return None

        if data.data.ndim > 1:
            raise ValueError("Data must be 1D")

        peaks = None
        rate = None
        if self.params["cardiorespiratory"]["input_type"].value == "ppg":
            datatype = "HRV"
            ppg, info = self.neurokit.ppg_process(data.data, sampling_rate=data.meta["sfreq"])
            rate = ppg["PPG_Rate"]
            variability_df = self.neurokit.hrv(info, sampling_rate=data.meta["sfreq"])
            peaks = ppg["PPG_Peaks"]

        elif self.params["cardiorespiratory"]["input_type"].value == "ecg":
            datatype = "HRV"
            # extract peaks
            ecg, info = self.neurokit.ecg_process(data.data, sampling_rate=data.meta["sfreq"])
            rate = ecg["ECG_Rate"]
            # compute hrv
            variability_df = self.neurokit.hrv(info, sampling_rate=data.meta["sfreq"])
            peaks = None

        elif self.params["cardiorespiratory"]["input_type"].value == "rsp":
            datatype = "RRV"
            rsp, info = self.neurokit.rsp_process(data.data, sampling_rate=data.meta["sfreq"])
            rate = rsp["RSP_Rate"]
            variability_df = self.neurokit.rsp_rrv(rsp, sampling_rate=data.meta["sfreq"])
            peaks = None

        BBorNN = "BB" if datatype == "RRV" else "NN"

        return {
            "Mean": (np.array(variability_df[f"{datatype}_Mean{BBorNN}"]), {}),
            "SDNN": (np.array(variability_df[f"{datatype}_SD{BBorNN}"]), {}),
            "SDSD": (np.array(variability_df[f"{datatype}_SDSD"]), {}),
            "RMSSD": (np.array(variability_df[f"{datatype}_RMSSD"]), {}),
            "VLF": (np.array(variability_df[f"{datatype}_VLF"]), {}),
            "LF": (np.array(variability_df[f"{datatype}_LF"]), {}),
            "HF": (np.array(variability_df[f"{datatype}_HF"]), {}),
            "LF/HF": (np.array(variability_df[f"{datatype}_LFHF"]), {}),
            #"LZC": (np.array(variability_df[f"{datatype}_LZC"]), {} if datatype == "HRV" else None),
            "Peaks": (np.array(peaks), {}),
            "Rate": (np.array(rate), {}) 
        }



import time
from multiprocessing import Manager as MPManager

import pytest
import yaml

from goofi.connection import Connection
from goofi.data import DataType
from goofi.message import Message, MessageType
from goofi.node import Node
from goofi.node_helpers import InputSlot, NodeRef, OutputSlot
from goofi.params import DEFAULT_PARAMS, NodeParams

from .utils import (
    DummyNode,
    FullDummyNode,
    ProcessingErrorNode,
    list_param_types,
    make_custom_node,
)

try:
    mp_manager = MPManager()
    Connection.set_backend("mp", mp_manager)
except AssertionError:
    # connection backend is already set
    pass


def test_abstract_node():
    # instantiating an abstract node should raise a TypeError
    with pytest.raises(TypeError):
        Node(Connection.create()[0])


def test_create_node():
    ref, n = DummyNode.create_local()

    # some basic checks
    assert n.alive, "Node is not alive."
    assert len(n.input_slots) == 0, "Node input slots are not empty."
    assert len(n.output_slots) == 0, "Node output slots are not empty."
    assert n.params == NodeParams(DEFAULT_PARAMS), "Node params are not default."
    assert n.messaging_thread.is_alive(), "Node messaging thread is not alive."
    assert n.processing_thread.is_alive(), "Node processing thread is not alive."

    # clean up dangling threads
    ref.terminate()


@pytest.mark.skip("This test is inconsistent")
def test_dead_pipe():
    ref, n = DummyNode.create_local()

    # close the connection
    ref.connection.close()
    time.sleep(0.05)

    # the node should stop its messaging thread and exit
    assert not n.messaging_thread.is_alive(), "Node messaging thread is alive."


def test_ping_pong():
    messages = []

    def callback(_: NodeRef, msg: Message):
        messages.append(msg)

    ref, _ = DummyNode.create_local()
    ref.set_message_handler(MessageType.PONG, callback)

    ref.connection.send(Message(MessageType.PING, {}))
    time.sleep(0.01)
    assert len(messages) == 1, "Node should have received one message."
    assert messages[0].type == MessageType.PONG, "Node should have received a pong message."

    # clean up dangling threads
    ref.terminate()


def test_terminate_message():
    ref, n = DummyNode.create_local()
    ref.connection.send(Message(MessageType.TERMINATE, {}))
    time.sleep(0.01)
    assert not n.alive, "Node should be dead after receiving terminate message."


def test_terminate_func():
    ref, n = DummyNode.create_local()
    ref.terminate()
    time.sleep(0.01)
    assert not n.alive, "Node should be dead after receiving terminate message."


def test_multiproc():
    ref = DummyNode.create()

    time.sleep(0.1)  # TODO: 0.1 is quite long but required on Windows, Linux works with 0.01
    # make sure the node is not dead
    assert ref.process.is_alive(), "Process should be alive."

    # send a terminate message
    ref.connection.send(Message(MessageType.TERMINATE, {}))
    time.sleep(0.1)  # TODO: 0.1 is quite long but required on Windows, Linux works with 0.01
    # make sure the node is dead
    assert not ref.process.is_alive(), "Process should be dead."


def test_processing_error(timeout: float = 1.5):
    messages = []

    def error_callback(ref: NodeRef, msg: Message):
        messages.append(msg)

    ref, n = ProcessingErrorNode.create_local()

    ref.set_message_handler(MessageType.PROCESSING_ERROR, error_callback)

    # manually trigger processing once
    n.process_flag.set()

    start = time.time()
    while len(messages) == 0:
        if time.time() - start > timeout:
            raise TimeoutError("Timeout while waiting for error message.")
        time.sleep(0.01)

    # make sure we received the error message
    assert len(messages) == 1, "Node should have received one message."

    # the processing thread should be alive as the node is expected to handle the error
    assert n.processing_thread.is_alive(), "Processing thread should still be alive."

    # the node and messaging thread should be alive
    assert n.alive, "Node should be alive."
    assert n.messaging_thread.is_alive(), "Messaging thread should be alive."

    # clean up dangling threads
    ref.terminate()


def test_full_node_input_slots():
    ref, n = FullDummyNode.create_local()

    for name, slot in ref.input_slots.items():
        assert isinstance(slot, DataType), f"NodeRef slots should be DataTypes, got {type(slot)}"
        assert name in n.input_slots, f"NodeRef input slot {name} is missing from node input slots."
        assert isinstance(
            n.input_slots[name], InputSlot
        ), f"Node input slot {name} should be an InputSlot, got {type(n.input_slots[name])}"
        assert slot == n.input_slots[name].dtype, f"Mismatching slot types for {name}: {slot} != {n.input_slots[name].dtype}"

    ref.terminate()


def test_full_node_output_slots():
    ref, n = FullDummyNode.create_local()

    for name, slot in ref.output_slots.items():
        assert isinstance(slot, DataType), f"NodeRef slots should be DataTypes, got {type(slot)}"
        assert name in n.output_slots, f"NodeRef output slot {name} is missing from node output slots."
        assert isinstance(
            n.output_slots[name], OutputSlot
        ), f"Node output slot {name} should be an OutputSlot, got {type(n.output_slots[name])}"
        assert slot == n.output_slots[name].dtype, f"Mismatching slot types for {name}: {slot} != {n.output_slots[name].dtype}"

    ref.terminate()


def test_full_node_params():
    ref, n = FullDummyNode.create_local()

    assert isinstance(ref.params, NodeParams), f"NodeRef params should be a NodeParams, got {type(ref.params)}"
    assert ref.params == n.params, "NodeRef params should be equal to node params."

    assert "common" in ref.params, "Node params should contain common params."
    assert "test" in ref.params, "Node params should contain test params."

    for param_type in list_param_types():
        name = "param_" + param_type.__name__.lower()
        assert name in ref.params.test, f"Node params should contain param '{name}'. Got {ref.params.test}"
        assert isinstance(
            getattr(ref.params.test, name), param_type
        ), f"Node param '{name}' should be a {param_type}, got {type(ref.params.test[name])}"

    ref.terminate()


def test_pipes():
    results = []

    def callback(**kwargs):
        results.append(kwargs["in"])

    cls1 = make_custom_node(output_slots={"out": DataType.ARRAY}, params={"common": {"autotrigger": True}})
    cls2 = make_custom_node(input_slots={"in": DataType.ARRAY}, process_callback=callback)

    ref1, n1 = cls1.create_local()
    ref2, _ = cls2.create_local()

    # connect the nodes
    ref1.connection.send(
        Message(
            MessageType.ADD_OUTPUT_PIPE,
            {"slot_name_out": "out", "slot_name_in": "in", "node_connection": ref2.connection},
        )
    )

    time.sleep(0.1)

    assert (
        len(n1.output_slots["out"].connections) == 2
    ), f"Expected two output slot connections, got {len(n1.output_slots['out'].connections)}"
    assert ("in", ref2.connection, False) in n1.output_slots["out"].connections, "Output slot connections are incorrect."
    assert len(results) > 0, "Processing callback should have been called once."

    # disconnect the nodes
    ref1.connection.send(
        Message(
            MessageType.REMOVE_OUTPUT_PIPE,
            {"slot_name_out": "out", "slot_name_in": "in", "node_connection": ref2.connection},
        )
    )

    time.sleep(0.01)

    assert len(n1.output_slots["out"].connections) == 1, "Output slot connections are incorrect."

    ref1.terminate()
    ref2.terminate()


@pytest.mark.parametrize("value", [10.0, 100.0])
def test_change_parameter(value):
    ref, n = DummyNode.create_local()

    # change a parameter
    ref.update_param("common", "max_frequency", value)

    time.sleep(0.01)

    # check that the parameter was changed
    assert n.params.common.max_frequency.value == value, "Parameter was not changed."

    # clean up dangling threads
    ref.terminate()


@pytest.mark.parametrize("value", [10.0, 100.0])
def test_change_parameter_callback(value):
    results = []

    def callback(_, value):
        results.append(value)

    # create a dummy node type with a callback for the common.max_frequency parameter
    CallbackDummyNode = type("CallbackDummyNode", (DummyNode,), {"common_max_frequency_changed": callback})
    ref, _ = CallbackDummyNode.create_local()

    # change a parameter
    ref.update_param("common", "max_frequency", value)

    time.sleep(0.01)

    # check that the parameter callback was called with the correct value
    assert len(results) == 1, "Parameter callback was not called."
    assert results[0] == value, "Parameter callback was called with the wrong value."

    # clean up dangling threads
    ref.terminate()


def test_serialize():
    result = None

    def callback(_, msg):
        nonlocal result
        result = msg.content

    ref, n = FullDummyNode.create_local()
    ref.set_message_handler(MessageType.SERIALIZE_RESPONSE, callback)

    # serialize the node
    ref.connection.send(Message(MessageType.SERIALIZE_REQUEST, {}))
    time.sleep(0.01)

    # "ground truth" serialized node
    node_type = "FullDummyNode"
    category = "tests"
    out_conns = {name: slot.connections for name, slot in n.output_slots.items()}
    params = n.params.serialize()

    # compare to the serialized node
    assert result["_type"] == node_type, "Node type is not serialized correctly."
    assert result["category"] == category, "Node category is not serialized correctly."
    assert result["out_conns"] == out_conns, "Output slots are not serialized correctly."
    assert result["params"] == params, "Params are not serialized correctly."

    # clean up dangling threads
    ref.terminate()


def test_serialize_yaml():
    result = None

    def callback(_, msg):
        nonlocal result
        result = msg.content

    ref, _ = FullDummyNode.create_local()
    ref.set_message_handler(MessageType.SERIALIZE_RESPONSE, callback)

    # serialize the node
    ref.connection.send(Message(MessageType.SERIALIZE_REQUEST, {}))
    time.sleep(0.01)

    # make sure the serialized parameters can be converted to YAML (except out_conns, which are removed in the manager)
    result.pop("out_conns")
    serialized = yaml.dump(result)
    reconstructed = yaml.load(serialized, Loader=yaml.FullLoader)

    assert reconstructed == result, "Serialized parameters are not valid YAML."

    # clean up dangling threads
    ref.terminate()



import time
from copy import deepcopy

import dearpygui.dearpygui as dpg

from goofi.node_helpers import list_nodes


def is_click_inside(item):
    """Check if the mouse click was inside the given item."""
    mouse_pos = dpg.get_mouse_pos(local=False)
    try:
        item_pos = dpg.get_item_state(item)["rect_min"]
    except KeyError:
        item_pos = dpg.get_item_pos(item)

    item_size = dpg.get_item_rect_size(item)
    return (
        mouse_pos[0] >= item_pos[0]
        and mouse_pos[1] >= item_pos[1]
        and mouse_pos[0] <= item_pos[0] + item_size[0]
        and mouse_pos[1] <= item_pos[1] + item_size[1]
    )


################################
######### Mouse Events #########
################################


def click_callback(_, btn, win):
    """Update node selection after a click event."""
    selected = dpg.get_selected_nodes(win.node_editor)
    if len(selected) == 1:
        win._select_node(selected[0])
    else:
        win._select_node(None)

    if win.create_node_window is not None and not is_click_inside(win.create_node_window):
        # the create node window is open but the click was outside of it, close it
        dpg.delete_item(win.create_node_window)
        win.create_node_window = None
    if win.node_info_window is not None and not is_click_inside(win.node_info_window):
        # the node info window is open but the click was outside of it, close it
        dpg.delete_item(win.node_info_window)
        win.node_info_window = None
    # TODO: do the same thing for the file selection window, which has a broken get_item_pos

    if btn == 1:
        for node in win.nodes.values():
            if is_click_inside(node.item):
                # right click on a node, open an information window
                node.display_info(win)
                break


def double_click_callback(_, btn, win):
    """If double click was left mouse button, open the window for creating a node."""
    if btn == 0 and len(dpg.get_selected_nodes(win.node_editor)) == 0:
        create_node(win)


################################
########## Key Events ##########
################################


def key_release_callback(_, data, win):
    """Handle key press events by calling the appropriate function from the key handler map."""
    if data in KEY_HANDLER_MAP:
        KEY_HANDLER_MAP[data](win)


#############################
########## Actions ##########
#############################


def delete_selected_item(win):
    """Deletes the selected node or link."""
    if any([dpg.is_item_active(item) for item in win.param_input_fields if dpg.does_item_exist(item)]):
        # an input field is active, do not delete the selected node
        return

    for node in dpg.get_selected_nodes(win.node_editor):
        win._remove_node(node)
    for link in dpg.get_selected_links(win.node_editor):
        win._remove_link(link)


def select_node_callback(_1, _2, user_data):
    """Callback for when a node is selected in the create node window."""
    win, node = user_data
    # clean up the state of the GUI
    escape(win)
    dpg.clear_selected_nodes(win.node_editor)
    dpg.clear_selected_links(win.node_editor)
    # create the node inside the manager, which will notify the window
    win.manager.add_node(node.__name__, node.category())


def create_selected_node(win):
    """Callback for when a node is selected in the create node window."""
    if win.create_node_window is None:
        return

    tab_bar, search_group, searchbox = dpg.get_item_user_data(win.create_node_window)
    if dpg.get_item_configuration(tab_bar)["show"]:
        # the tab bar is visible, the user has not selected a node yet
        return

    # get the selected node
    try:
        top_btn = dpg.get_item_children(search_group)[1][0]
    except IndexError:
        # no node was selected
        dpg.focus_item(searchbox)
        return

    # "click" the button
    user_data = dpg.get_item_user_data(top_btn)
    dpg.get_item_configuration(top_btn)["callback"](top_btn, None, user_data)


def create_node(win):
    """Opens the window for creating a node. If the window is already open, switch to the next tab."""
    # create a dictionary of nodes by category
    categories = {}
    for node in list_nodes():
        cat = node.category()
        if cat not in categories:
            categories[cat] = []
        categories[cat].append(node)

    if win.create_node_window is not None:
        if dpg.is_item_focused(win.create_node_window):
            # the window is open and focused, switch to the next tab
            tab_bar = dpg.get_item_user_data(win.create_node_window)[0]
            # increment tab index
            if dpg.is_key_down(dpg.mvKey_Shift):
                win.last_create_node_tab = (win.last_create_node_tab - 1) % len(categories)
            else:
                win.last_create_node_tab = (win.last_create_node_tab + 1) % len(categories)
            # switch to the next tab
            tab = f"tab_{list(categories.keys())[win.last_create_node_tab]}"
            dpg.set_value(tab_bar, tab)
            return

        # the window is open but not focused, close it to reopen later in this function
        dpg.delete_item(win.create_node_window)
        win.create_node_window = None

    if not dpg.is_item_hovered(win.node_editor):
        # the node editor is not hovered, do not open the window
        return

    # create a new window instance
    win.create_node_window = dpg.add_window(
        label="Create Node", pos=dpg.get_mouse_pos(local=False), no_collapse=True, autosize=True
    )

    def search_callback(_, data):
        """Callback for when the search bar changes."""
        tab_bar, search_group, _ = dpg.get_item_user_data(win.create_node_window)

        if len(data) == 0:
            # no search query, show all nodes
            dpg.configure_item(tab_bar, show=True)
            dpg.configure_item(search_group, show=False)
            return

        # hide all tabs
        dpg.configure_item(tab_bar, show=False)
        # clear the search group
        dpg.delete_item(search_group, children_only=True)

        # add a button for each node that matches the search query
        for node in list_nodes():
            if data.lower() in node.__name__.lower():
                dpg.add_button(label=node.__name__, callback=select_node_callback, user_data=(win, node), parent=search_group)

        # show the search group
        dpg.configure_item(search_group, show=True)

    # add search bar
    searchbox = dpg.add_input_text(hint="search", parent=win.create_node_window, callback=search_callback)
    win.param_input_fields.append(searchbox)
    dpg.add_separator(parent=win.create_node_window)

    # create a tab bar with a tab for each category
    with dpg.tab_bar(parent=win.create_node_window) as tab_bar:
        for cat, nodes in categories.items():
            with dpg.tab(label=cat, tag=f"tab_{cat}"):
                # create a button for each node in the category
                for node in nodes:
                    dpg.add_button(label=node.__name__, callback=select_node_callback, user_data=(win, node))

    # create a vertical group for listing nodes during search
    search_group = dpg.add_group(horizontal=False, parent=win.create_node_window, show=False)

    # switch to the current tab
    dpg.set_value(tab_bar, f"tab_{list(categories.keys())[win.last_create_node_tab]}")
    # store the tab bar and search group in the window user data
    dpg.set_item_user_data(win.create_node_window, (tab_bar, search_group, searchbox))

    # focus the search bar
    dpg.focus_item(searchbox)


def escape(win):
    """
    Provides a way to cancel the creation of a node by closing the node selection window.
    If it was already closed, clears the current node and link selection.
    """
    if win.create_node_window is not None:
        # close the create node window
        dpg.delete_item(win.create_node_window)
        win.create_node_window = None
    elif win.file_selection_window is not None:
        # close the file selection window
        dpg.delete_item(win.file_selection_window)
        win.file_selection_window = None
    else:
        # clear the node and link selection
        dpg.clear_selected_nodes(win.node_editor)
        dpg.clear_selected_links(win.node_editor)

    if win.node_info_window is not None:
        # close the node info window
        dpg.delete_item(win.node_info_window)
        win.node_info_window = None


def save_manager(win):
    """Save the manager to a file. We go through the Window to potentially open a file selection dialog."""
    if dpg.is_key_down(dpg.mvKey_Control):
        win.save()


def copy_selected_nodes(win, timeout: float = 0.1):
    """Copy the selected nodes to the clipboard."""
    if not dpg.is_key_down(dpg.mvKey_Control):
        return

    # retrieve selected nodes and their positions
    nodes = {n: dpg.get_item_user_data(n) for n in dpg.get_selected_nodes(win.node_editor)}
    if len(nodes) == 0:
        return

    positions = [dpg.get_item_pos(n) for n in dpg.get_selected_nodes(win.node_editor)]
    avg_pos = [sum(p[0] for p in positions) / len(positions), sum(p[1] for p in positions) / len(positions)]

    # wait for all nodes to respond, if their serialization_pending flag is set
    start = time.time()
    serialized_nodes = []
    for (item, node), pos in zip(nodes.items(), positions):
        while node.serialization_pending and time.time() - start < timeout:
            # wait for the node to respond or for the timeout to be reached
            time.sleep(0.01)

        # get node name
        names = [name for name, gui_node in win.nodes.items() if gui_node.item == item]
        assert len(names) == 1, f"The following nodes seem to be duplicates: {names}"
        name = names[0]

        # check if we got a response in time
        if node.serialization_pending:
            # TODO: add proper logging
            print(f"WARNING: Node {name} timed out while waiting for serialization. Node state is possibly outdated.")

        if node.serialized_state is None:
            # TODO: add proper logging
            print(f"ERROR: Node {name} does not have a serialized state. Copying is not possible.")
            win.node_clipboard = None
            return

        # get links connected to input slots of the current node
        input_links = []
        for n1, n2, s1, s2 in win.links.keys():
            if n2 == name:
                input_links.append((n1, n2, s1, s2))

        # store the serialized state
        ser = deepcopy(node.serialized_state)
        ser["gui_kwargs"] = {"offset": [pos[0] - avg_pos[0], pos[1] - avg_pos[1]]}
        ser["name"] = name
        ser["input_links"] = input_links
        # we don't need the output connections
        del ser["out_conns"]
        serialized_nodes.append(ser)

    # store the serialized nodes in the clipboard
    win.node_clipboard = serialized_nodes


def paste_nodes(win):
    """Paste the nodes from the clipboard."""
    if not dpg.is_key_down(dpg.mvKey_Control) or win.node_clipboard is None:
        return

    # add the nodes to the manager
    rename_nodes = {}
    for node in win.node_clipboard:
        new_name = win.manager.add_node(node["_type"], node["category"], params=node["params"], **node["gui_kwargs"])
        rename_nodes[node["name"]] = new_name

    # add links between the nodes
    for node in win.node_clipboard:
        for n1, n2, s1, s2 in node["input_links"]:
            n1 = rename_nodes[n1] if n1 in rename_nodes else n1
            n2 = rename_nodes[n2] if n2 in rename_nodes else n2
            win.manager.add_link(n1, n2, s1, s2)


# the key handler map maps key press events to functions that handle them
KEY_HANDLER_MAP = {
    dpg.mvKey_Delete: delete_selected_item,
    dpg.mvKey_Back: delete_selected_item,
    dpg.mvKey_Tab: create_node,
    dpg.mvKey_Escape: escape,
    dpg.mvKey_S: save_manager,
    dpg.mvKey_C: copy_selected_nodes,
    dpg.mvKey_V: paste_nodes,
    dpg.mvKey_Return: create_selected_node,
}



import pytest

from goofi.data import DTYPE_TO_TYPE, Data

from .utils import list_data_types


@pytest.mark.parametrize("dtype", list_data_types())
def test_create_data(dtype):
    # all dtype checks should pass
    Data(dtype, dtype.empty(), {})

    # data is None, should raise ValueError
    with pytest.raises(ValueError):
        Data(dtype, None, {})

    # metadata is None, should raise ValueError
    with pytest.raises(ValueError):
        Data(dtype, dtype.empty(), None)

    # dtype is None, should raise ValueError
    with pytest.raises(ValueError):
        Data(None, dtype.empty(), {})

    # make sure all other dtypes raise a ValueError
    for other_dtype in list_data_types():
        if other_dtype == dtype:
            continue
        with pytest.raises(ValueError):
            Data(dtype, other_dtype.empty(), {})


@pytest.mark.parametrize("dtype", list_data_types())
def test_dtype_map(dtype):
    assert dtype in DTYPE_TO_TYPE, f"Missing entry in DTYPE_TO_TYPE for dtype {dtype}."



import numpy as np

from goofi.data import Data, DataType
from goofi.node import Node
from goofi.params import BoolParam


class WelfordsZTransform(Node):
    INIT_STEPS = 50

    def config_input_slots():
        return {"data": DataType.ARRAY}

    def config_output_slots():
        return {"normalized": DataType.ARRAY}

    def config_params():
        return {"welford": {"biased_std": False, "outlier_stds": 4.0, "reset": BoolParam(False, trigger=True)}}

    def setup(self):
        self.mean = {}
        self.m2 = {}

    def process(self, data: Data):
        if data is None or data.data is None:
            return None

        val = np.asarray(data.data)  # use asarray instead of squeeze to maintain dimensionality
        if val.ndim > 2:
            print("Error: WelfordsZTransform only accepts 1D or 2D arrays")
            return None

        normalized_value = np.zeros_like(val)

        if self.params["welford"]["reset"].value is True:
            self.reset()
        if val.ndim == 1:
            iterator = enumerate(val)
        else:  # if 2D array, apply the transform along the last dimension
            iterator = np.ndenumerate(val)

        for idx, single_val in iterator:
            key = str(idx)
            if key not in self.mean:
                self.mean[key] = {"mean": 0.0, "count": 0}
                self.m2[key] = 0.0

            count = self.mean[key]["count"]
            biased = self.params["welford"]["biased_std"].value
            outlier_stds = self.params["welford"]["outlier_stds"].value

            delta = single_val - self.mean[key]["mean"]
            count += 1
            new_mean = self.mean[key]["mean"] + delta / count
            delta2 = single_val - new_mean
            self.m2[key] += delta * delta2

            if count < self.INIT_STEPS:
                if val.ndim == 1:
                    normalized_value[idx] = 0
                else:
                    normalized_value[idx[0], idx[1]] = 0
            else:
                std_dev = np.sqrt(self.m2[key] / (count if biased else count - 1) + 1e-8)
                if abs(delta) < outlier_stds * std_dev:
                    if val.ndim == 1:
                        normalized_value[idx] = delta / std_dev
                    else:
                        normalized_value[idx[0], idx[1]] = delta / std_dev
                else:
                    if val.ndim == 1:
                        normalized_value[idx] = outlier_stds * np.sign(delta)
                    else:
                        normalized_value[idx[0], idx[1]] = outlier_stds * np.sign(delta)

            self.mean[key] = {"mean": new_mean, "count": count}

        return {"normalized": (normalized_value, data.meta)}

    def reset(self):
        self.mean = {}
        self.m2 = {}



import numpy as np
from goofi.data import Data, DataType
from goofi.node import Node


class Compass(Node):
    """
    A Goofi node that calculates angles from polar arrays in N-dimensional space.
    Takes two inputs (pole1, pole2) and computes the angles corresponding to N-1 dimensions.
    """

    def config_input_slots():
        return {"pole1": DataType.ARRAY, "pole2": DataType.ARRAY}

    def config_output_slots():
        return {"angles": DataType.ARRAY}

    def process(self, pole1: Data, pole2: Data):
        """
        Processes the input polar arrays to compute angles in N-dimensional space.

        Args:
            pole1 (Data): First input array (1D, length N).
            pole2 (Data): Second input array (1D, length N).

        Returns:
            dict: Angles in N-1 dimensions.
        """
        if pole1 is None or pole2 is None:
            return None

        # Extract the float values from Data objects
        pole1_val = pole1.data
        pole2_val = pole2.data

        # Ensure the inputs are valid 1D arrays of the same length
        if pole1_val.ndim != 1 or pole2_val.ndim != 1 or len(pole1_val) != len(pole2_val):
            raise ValueError("Inputs pole1 and pole2 must be 1D arrays of the same length.")

        # Calculate the vector difference
        vector_diff = pole2_val - pole1_val

        # Initialize angles array
        angles = []

        # Compute angles iteratively
        for i in range(len(vector_diff) - 1):
            y = vector_diff[i + 1]
            x = vector_diff[i]
            angle = np.arctan2(y, x)  # Use arctan2 for full circle coverage
            angles.append(angle)

        # Convert angles to degrees and normalize to [0, 360)
        angles_degrees = np.degrees(angles)
        angles_degrees = np.mod(angles_degrees, 360)  # Ensure values are in [0, 360)

        # Return the angles as a Data object
        return {"angles": (np.array(angles_degrees), {})}



import platform
import time
from os import path

import pytest

import goofi
from goofi.connection import Connection
from goofi.manager import Manager
from goofi.message import Message, MessageType

MANAGER_TEST_DURATION = 0.1


def create_simple_manager(comm_backend: str = "mp") -> Manager:
    """
    Creates a simple manager with a constant node, a sine node, and an add node.

    ### Parameters
    `comm_backend` : str
        The communication backend to use. Choose from "mp", "zmq-tcp" or "zmq-ipc".

    ### Returns
    manager : Manager
        The manager object.
    """
    manager = Manager(duration=MANAGER_TEST_DURATION, communication_backend=comm_backend)
    manager.add_node("ConstantArray", "inputs")
    manager.add_node("Sine", "inputs")
    manager.add_node("Operation", "array")

    manager.add_link("constantarray0", "operation0", "out", "a")
    manager.add_link("sine0", "operation0", "out", "b")
    return manager


def test_creation():
    Manager(duration=MANAGER_TEST_DURATION)


def test_main():
    goofi.manager.main(MANAGER_TEST_DURATION, ["--headless"])


@pytest.mark.skipif(platform.system() == "Windows", reason="Multiprocessing is very slow on Windows.")
@pytest.mark.parametrize("comm_backend", Connection.get_backends().keys())
def test_simple(comm_backend):
    if comm_backend.startswith("zmq"):
        # TODO: make sure zmq backend works
        pytest.skip("ZeroMQ backend still has some issues.")

    manager = create_simple_manager(comm_backend=comm_backend)
    my_conn, node_conn = Connection.create()
    manager.nodes["operation0"].connection.send(
        Message(MessageType.ADD_OUTPUT_PIPE, {"slot_name_out": "out", "slot_name_in": "in", "node_connection": my_conn})
    )

    last = None
    rates = []
    data = []
    for _ in range(10):
        msg = node_conn.recv()
        data.append(msg.content["data"].data[0])

        if last is not None:
            rates.append(1 / (time.time() - last))
        last = time.time()

    # data should be between 0 and 2
    assert all(0 <= x <= 2 for x in data), "Data should be between 0 and 2."

    # mean rate should be ~30 Hz
    mean_rate = sum(rates) / len(rates)
    assert mean_rate == pytest.approx(30, abs=0.5), f"Mean rate should be ~30 Hz, got {mean_rate} Hz."

    # clean up
    manager.terminate()


def test_save_empty(tmpdir):
    manager = Manager(duration=MANAGER_TEST_DURATION)

    # if path is a file, save to that file
    manager.save(path.join(tmpdir, "test.gfi"))
    assert path.exists(path.join(tmpdir, "test.gfi")), "Expected file test.gfi to exist."

    # clean up
    manager.terminate()


def test_save_extension(tmpdir):
    manager = create_simple_manager()

    # make sure the file gets the correct extension
    manager.save(path.join(tmpdir, "test2"))
    assert path.exists(path.join(tmpdir, "test2.gfi")), "Expected file extension to be set to .gfi"

    # clean up
    manager.terminate()


@pytest.mark.parametrize("overwrite", [True, False])
def test_save_simple(overwrite, tmpdir):
    manager = create_simple_manager()

    tmpdir = str(tmpdir)

    # if path is a file, save to that file
    manager.save(path.join(tmpdir, "test.gfi"), overwrite=overwrite)
    assert path.exists(path.join(tmpdir, "test.gfi")), "Expected file test.gfi to exist."

    time.sleep(0.1)

    # if file already exists, raise FileExistsError
    if overwrite:
        manager.save(path.join(tmpdir, "test.gfi"), overwrite=True)
        assert path.exists(path.join(tmpdir, "test.gfi")), "test.gfi should still be there."
    else:
        with pytest.raises(FileExistsError):
            # overwrite is False by default
            manager.save(path.join(tmpdir, "test.gfi"))

    # clean up
    manager.terminate()


def test_save_simple_dir(tmpdir):
    manager = create_simple_manager()

    with pytest.raises(ValueError):
        # if path is not a string, raise an error (tmpdir is a Path object)
        manager.save(tmpdir)

    tmpdir = str(tmpdir)

    # if path is a directory, save to untitled0.gfi
    manager.save(tmpdir)
    assert path.exists(path.join(tmpdir, "untitled0.gfi")), "Expected file untitled0.gfi to exist."

    time.sleep(0.1)

    # if untitled0.gfi exists in the directory, save to untitled1.gfi
    manager.save(tmpdir)
    assert path.exists(path.join(tmpdir, "untitled1.gfi")), "Expected file untitled1.gfi to exist."

    # clean up
    manager.terminate()



import importlib.resources as pkg_resources
import time
import traceback
from abc import ABC, abstractmethod
from copy import deepcopy
from enum import Enum
from multiprocessing import Process
from os.path import dirname, join
from pathlib import PosixPath
from threading import Event, Thread
from typing import Any, Callable, Dict, Optional, Tuple, Union

from goofi import assets
from goofi.connection import Connection
from goofi.data import Data, DataType
from goofi.message import Message, MessageType
from goofi.node_helpers import InputSlot, NodeRef, OutputSlot
from goofi.params import InvalidParamError, NodeParams


class MultiprocessingForbiddenError(Exception):
    pass


def require_init(func: Callable) -> Callable:
    """
    Decorator that checks if `super().__init__()` has been called in the `__init__()` method of the class
    that the decorated method belongs to. This is used to make sure that the base class is initialized before
    accessing any of its attributes.

    ### Parameters
    `func` : Callable
        The method to decorate.

    ### Returns
    `Callable`
        The decorated method.
    """

    def wrapper(self, *args, **kwargs):
        # check if the base class is initialized
        if not hasattr(self, "_alive"):
            raise RuntimeError("Make sure to call super().__init__() in your node's __init__ method.")
        return func(self, *args, **kwargs)

    return wrapper


class NodeEnv(Enum):
    """
    Enumeration of the possible node environments.
    """

    MULTIPROCESSING = 1  # the node owns its own process
    LOCAL = 2  # the node is running in the manager's process
    STANDALONE = 3  # running in the main process, but without a manager


class Node(ABC):
    """
    The base class for all nodes. A node is a processing unit that can receive data from other nodes, process the
    data, and send the processed data to other nodes. A node can have any number of input and output slots. Each
    input slot can receive data from one output slot, and each output slot can send data to any number of input
    slots.

    ### Parameters
    `connection` : Connection
        The input connection to the node. This is used to receive messages from the manager, or other nodes.
    `input_slots` : Dict[str, InputSlot]
        A dict containing the input slots of the node. The keys are the names of the input slots, and the values
        are the input slots themselves.
    `output_slots` : Dict[str, OutputSlot]
        A dict containing the output slots of the node. The keys are the names of the output slots, and the values
        are the output slots themselves.
    `params` : NodeParams
        An instance of the NodeParams class containing the parameters of the node.
    `environment` : NodeEnv
        The environment in which the node is running.
    """

    NO_MULTIPROCESSING = False
    MESSAGE_TIMEOUT = 500  # ms

    def __init__(
        self,
        connection: Optional[Connection],
        input_slots: Dict[str, InputSlot],
        output_slots: Dict[str, OutputSlot],
        params: NodeParams,
        environment: NodeEnv,
    ) -> None:
        # initialize the base class
        self._alive = True
        self._node_ready = False

        # store the arguments and validate them
        self.connection = connection
        self._input_slots = input_slots
        self._output_slots = output_slots
        self._params = params

        self._validate_attrs(environment)

        # initialize node flags
        self.process_flag = Event()
        if self.params.common.autotrigger.value:
            self.process_flag.set()

        # set up dict of possibly timed out output connections
        self.pending_connections = {}

        if environment != NodeEnv.STANDALONE:
            # initialize data processing thread
            self.processing_thread = Thread(target=self._processing_loop, daemon=True)
            self.processing_thread.start()

        if environment == NodeEnv.MULTIPROCESSING:
            # this is a separate process, run the messaging loop in the current thread
            # NOTE: if we don't block the current thread, the node's process will die
            self._messaging_loop()
        elif environment == NodeEnv.LOCAL:
            # this is the main process, create a new thread to not block it
            self.messaging_thread = Thread(target=self._messaging_loop, daemon=True)
            self.messaging_thread.start()
        elif environment == NodeEnv.STANDALONE:
            # the node does not have a messaging, or processing loop when running in standalone mode
            pass
        else:
            raise ValueError(f"Invalid environment: {environment}")

    @require_init
    def _validate_attrs(self, environment: NodeEnv) -> None:
        """
        Check that all attributes are present and of the correct type.
        """
        # check connection type
        if environment == NodeEnv.STANDALONE:
            if self.connection is not None:
                raise ValueError("Running in standalone mode, connection should be None.")
        else:
            if not isinstance(self.connection, Connection):
                raise TypeError(f"Expected Connection, got {type(self.connection)}")

        # check input slots type
        for name, slot in self._input_slots.items():
            if not isinstance(name, str) or len(name) == 0:
                raise ValueError(f"Expected input slot name '{name}' to be a non-empty string.")
            if not isinstance(slot, InputSlot):
                raise TypeError(f"Expected InputSlot for input slot '{name}', got {type(slot)}")

        # check output slots type
        for name, slot in self._output_slots.items():
            if not isinstance(name, str) or len(name) == 0:
                raise ValueError(f"Expected output slot name '{name}' to be a non-empty string.")
            if not isinstance(slot, OutputSlot):
                raise TypeError(f"Expected OutputSlot for output slot '{name}', got {type(slot)}")

        # check params type
        if not isinstance(self._params, NodeParams):
            raise TypeError(f"Expected NodeParams, got {type(self._params)}")

    def _setup(self):
        """This method calls the node's setup method and handles any exceptions that may occur."""
        while not self._node_ready:
            try:
                self.setup()
                self._node_ready = True
                # clear any errors that may have occurred during setup
                self.connection.try_send(Message(MessageType.PROCESSING_ERROR, {"error": None}))
            except Exception:
                error_message = traceback.format_exc()
                self.connection.try_send(Message(MessageType.PROCESSING_ERROR, {"error": error_message}))
                time.sleep(0.1)

    def _serialize(self):
        """Serialize the node's type, output connections, and parameters, and send the serialized data to the manager."""
        # serialize output connections and the node's parameters
        out_conns = {name: slot.connections for name, slot in self.output_slots.items()}
        params = self.params.serialize()
        # return the serialized data
        self.connection.try_send(
            Message(
                MessageType.SERIALIZE_RESPONSE,
                {"_type": type(self).__name__, "category": self.category(), "out_conns": out_conns, "params": params},
            )
        )

    def _messaging_loop(self):
        """
        This method runs in a separate thread and handles incoming messages from the manager, or other nodes.
        """
        # run the node's setup method
        if self._setup.__self__.__class__._setup is not Node._setup:
            # the node implements _setup, which is not allowed
            error = (
                f"The {self.__class__.__name__} node implements the _setup() method, which is reserved for "
                "internal use. Use setup() instead."
            )
            self.connection.try_send(Message(MessageType.PROCESSING_ERROR, {"error": error}))
            raise RuntimeError(error)

        Thread(target=self._setup, daemon=True).start()

        # run the messaging loop
        while self.alive:
            # receive the message
            try:
                msg = self.connection.recv()
            except ConnectionError:
                # the connection was closed, consider the node dead
                self._alive = False
                self.connection.close()
                continue
            except Exception:
                error_message = traceback.format_exc()
                self.connection.try_send(Message(MessageType.PROCESSING_ERROR, {"error": error_message}))
                continue

            # potentially restart the processing thread
            # TODO: this might be starting too many threads and not cleaning up properly
            if not self.processing_thread.is_alive():
                self.processing_thread = Thread(target=self._processing_loop, daemon=True)
                self.processing_thread.start()

            if not isinstance(msg, Message):
                raise TypeError(f"Expected Message, got {type(msg)}")

            if msg.type == MessageType.PING:
                # respond to a ping message by sending a pong message
                self.connection.try_send(Message(MessageType.PONG, {}))
            elif msg.type == MessageType.TERMINATE:
                # terminate the node
                self._alive = False
                # clear data in connected downstream nodes
                for slot in self.output_slots.values():
                    for slot_name, conn, _ in slot.connections:
                        conn.try_send(Message(MessageType.CLEAR_DATA, {"slot_name": slot_name}))
            elif msg.type == MessageType.ADD_OUTPUT_PIPE:
                # add a connection to the output slot
                slot = self.output_slots[msg.content["slot_name_out"]]
                conn = msg.content["node_connection"]
                self_conn = False

                if conn is None:
                    # if no connection is specified, connect to the node's own node reference
                    conn = self.connection
                    self_conn = True

                slot.connections.append((msg.content["slot_name_in"], conn, self_conn))

                if not self_conn:
                    # notify the manager that the connection was added
                    self._serialize()
            elif msg.type == MessageType.REMOVE_OUTPUT_PIPE:
                # clear the data in the input slot
                msg.content["node_connection"].try_send(
                    Message(MessageType.CLEAR_DATA, {"slot_name": msg.content["slot_name_in"]})
                )
                # remove the connection
                slot = self.output_slots[msg.content["slot_name_out"]]
                try:
                    slot.connections.remove((msg.content["slot_name_in"], msg.content["node_connection"], False))
                except ValueError:
                    # connection doesn't exist
                    self.connection.try_send(
                        Message(
                            MessageType.PROCESSING_ERROR,
                            {
                                "error": f"Request to remove non-existent connection from "
                                f"{msg.content['slot_name_out']} to {msg.content['slot_name_in']}."
                            },
                        )
                    )

                # notify the manager of the updated connections
                self._serialize()
            elif msg.type == MessageType.DATA:
                # received data from another node
                if msg.content["slot_name"] not in self.input_slots:
                    raise ValueError(f"Received DATA message but input slot '{msg.content['slot_name']}' doesn't exist.")
                slot = self.input_slots[msg.content["slot_name"]]
                slot.data = msg.content["data"]
                if slot.trigger_process:
                    self.process_flag.set()
            elif msg.type == MessageType.CLEAR_DATA:
                # clear the data in the input slot (usually triggered by a REMOVE_OUTPUT_PIPE message)
                slot = self.input_slots[msg.content["slot_name"]]
                slot.clear()
            elif msg.type == MessageType.PARAMETER_UPDATE:
                # update a parameter
                group = msg.content["group"]
                param_name = msg.content["param_name"]
                param_value = msg.content["param_value"]
                if group not in self.params:
                    raise ValueError(f"Parameter group '{group}' doesn't exist.")
                if param_name not in self.params[group]:
                    raise ValueError(f"Parameter '{param_name}' doesn't exist in group '{group}'.")
                self.params[group][param_name].value = param_value

                # notify the manager that the parameter was updated
                self._serialize()

                # call the callback if it exists
                if hasattr(self, f"{group}_{param_name}_changed"):
                    try:
                        getattr(self, f"{group}_{param_name}_changed")(param_value)
                    except Exception as e:
                        # parameter callback raised an exception, send out an error message
                        self.connection.try_send(
                            Message(
                                MessageType.PROCESSING_ERROR,
                                {"error": f"Parameter callback for {group}.{param_name} failed: {e}"},
                            )
                        )

            elif msg.type == MessageType.SERIALIZE_REQUEST:
                self._serialize()
            else:
                # TODO: handle the incoming message
                raise NotImplementedError(f"Message type {msg.type} not implemented.")

        # run the node's terminate method
        try:
            self.terminate()
        except Exception as e:
            self.connection.try_send(Message(MessageType.PROCESSING_ERROR, {"error": str(e)}))

        # close input connection
        self.connection.close()

    def _processing_loop(self):
        """
        This method runs in a separate thread and handles the processing of input data and sending of
        output data to other nodes.
        """
        # wait until the node's setup is complete
        while not self._node_ready:
            time.sleep(0.1)

        last_update = 0
        while self.alive:
            # wait for a trigger
            self.process_flag.wait()
            # clear the trigger if autotrigger is False
            if not self.params.common.autotrigger.value:
                self.process_flag.clear()

            # limit the update rate
            if self.params.common.max_frequency.value > 0:
                sleep_time = 1 / self.params.common.max_frequency.value - (time.time() - last_update)
                if sleep_time > 0:
                    time.sleep(sleep_time)
                last_update = time.time()

            # gather input data
            input_data = {name: slot.data for name, slot in self.input_slots.items()}

            try:
                # process data
                output_data = self.process(**input_data)
            except Exception:
                error_message = traceback.format_exc()
                self.connection.try_send(Message(MessageType.PROCESSING_ERROR, {"error": error_message}))
                continue

            if not self.alive:
                # the node was terminated during processing
                break

            # if process returns None, skip sending output data
            if output_data is None:
                # TODO: make sure this is the correct behavior
                continue

            # check that the process method returned a dict
            if not isinstance(output_data, dict):
                raise TypeError(
                    f"The process method should return a dictionary with one entry per output slot. "
                    f"Got {type(output_data)}."
                )

            # check that the output data contains the correct fields
            if missing := set(self.output_slots.keys()) - set(output_data.keys()):
                self.connection.try_send(
                    Message(
                        MessageType.PROCESSING_ERROR,
                        {
                            "error": f"Missing output fields: {missing}. "
                            f"Make sure that the process method returns a dict with one entry per output slot."
                        },
                    )
                )

            # check that the output data doesn't contain extra fields
            if extra_fields := list(set(output_data.keys()) - set(self.output_slots.keys())):
                self.connection.try_send(
                    Message(
                        MessageType.PROCESSING_ERROR,
                        {
                            "error": f"Extra output fields: {extra_fields}. "
                            f"The process method should only return those fields that were specified in the output slots."
                        },
                    )
                )

            # send output data
            for name in self.output_slots.keys():
                data = output_data[name]
                if data is None:
                    # skip sending None data
                    continue

                try:
                    data = Data(self.output_slots[name].dtype, data[0], data[1])
                except Exception:
                    error_message = traceback.format_exc()
                    self.connection.try_send(Message(MessageType.PROCESSING_ERROR, {"error": error_message}))
                    continue

                # send the data to all connected nodes
                for target_slot, conn, self_conn in self.output_slots[name].connections:
                    try:
                        msg = Message(MessageType.DATA, {"data": data, "slot_name": target_slot})
                    except Exception:
                        error_message = traceback.format_exc()
                        self.connection.try_send(Message(MessageType.PROCESSING_ERROR, {"error": error_message}))
                        continue

                    if conn._id in self.pending_connections:
                        # filter out dead threads
                        self.pending_connections[conn._id] = [
                            (thread, timestamp) for thread, timestamp in self.pending_connections[conn._id] if thread.is_alive()
                        ]
                        # check if the connection has timed out
                        timeout_occurred = False
                        for _, creation in self.pending_connections[conn._id]:
                            if time.time() - creation > self.MESSAGE_TIMEOUT / 1000:
                                # the connection has timed out, remove it
                                # TODO: figure out what's going on, if we remove a timed out connection the GUI will sometimes lose contact with the node
                                # self.output_slots[name].connections.remove((target_slot, conn, self_conn))
                                timeout_occurred = True
                                continue

                        if timeout_occurred:
                            # skip sending the message if the connection has timed out
                            continue
                    else:
                        self.pending_connections[conn._id] = []

                    # send the message (in a separate thread because connections may time out and block)
                    t = Thread(target=conn.send, args=(msg,), daemon=True)
                    t.start()
                    self.pending_connections[conn._id].append((t, time.time()))

    @staticmethod
    def _configure(cls) -> Tuple[Dict[str, InputSlot], Dict[str, OutputSlot], NodeParams]:
        """Retrieves the node's configuration of input slots, output slots, and parameters."""
        in_slots = cls.config_input_slots()
        out_slots = cls.config_output_slots()
        params = cls.config_params()

        return (
            {name: slot if isinstance(slot, InputSlot) else InputSlot(slot) for name, slot in in_slots.items()},
            {name: slot if isinstance(slot, OutputSlot) else OutputSlot(slot) for name, slot in out_slots.items()},
            NodeParams(params),
        )

    def common_autotrigger_changed(self, value):
        """If the new value of the parameter common.autotrigger is True, trigger the processing loop."""
        if value:
            self.process_flag.set()

    @require_init
    def clear_error(self):
        """Clear the error message."""
        self.connection.try_send(Message(MessageType.PROCESSING_ERROR, {"error": None}))

    @classmethod
    def create(cls, initial_params: Optional[Dict[str, Dict[str, Any]]] = None, retries: int = 3) -> NodeRef:
        """
        Create a new node instance in a separate process and return a reference to the node.

        ### Parameters
        `initial_params` : Optional[Dict[str, Dict[str, Any]]]
            A dictionary of parameter groups, where each group is a dictionary of parameter names and values.
            Defaults to None.

        ### Returns
        `NodeRef`
            A reference to the node.
        """
        if cls.NO_MULTIPROCESSING:
            raise MultiprocessingForbiddenError("Multiprocessing is forbidden for this node.")

        # generate arguments for the node
        in_slots, out_slots, params = cls._configure(cls)
        # integrate initial parameters if they are provided
        if initial_params is not None:
            try:
                params.update(initial_params)
            except InvalidParamError:
                print(
                    "\n====================== ERROR ======================\n"
                    f"Setting parameters failed for {cls.__name__}. This is likely due to updates"
                    " in the node's code. Make sure to reconfigure the node."
                    "\n===================================================\n"
                )

        tries = 0
        while True:
            try:
                conn1, conn2 = Connection.create()

                # instantiate the node in a separate process
                proc = Process(target=cls, args=(conn2, in_slots, out_slots, params, NodeEnv.MULTIPROCESSING), daemon=True)
                proc.start()
                break
            except Exception as e:
                tries += 1
                if tries >= retries:
                    raise e
                time.sleep(0.1)

        # create the node reference
        return NodeRef(
            conn1,
            {name: slot.dtype for name, slot in in_slots.items()},
            {name: slot.dtype for name, slot in out_slots.items()},
            params,
            cls.category(),
            process=proc,
        )

    @classmethod
    def create_local(cls, initial_params: Optional[Dict[str, Dict[str, Any]]] = None) -> Tuple[NodeRef, "Node"]:
        """
        Create a new node instance in the current process and return a reference to the node,
        as well as the node itself.

        ### Parameters
        `initial_params` : Optional[Dict[str, Dict[str, Any]]]
            A dictionary of parameter groups, where each group is a dictionary of parameter names and values.
            Defaults to None.

        ### Returns
        `Tuple[NodeRef, Node]`
            A tuple containing the node reference and the node itself.
        """
        # generate arguments for the node
        in_slots, out_slots, params = cls._configure(cls)

        # integrate initial parameters if they are provided
        if initial_params is not None:
            try:
                params.update(initial_params)
            except InvalidParamError:
                print(
                    "\n====================== ERROR ======================\n"
                    f"Setting parameters failed for {cls.__name__}. This is likely due to updates"
                    " in the node's code. Make sure to reconfigure the node."
                    "\n===================================================\n"
                )

        conn1, conn2 = Connection.create()
        # instantiate the node in the current process
        node = cls(conn2, in_slots, out_slots, params, NodeEnv.LOCAL)
        # create the node reference
        return (
            NodeRef(
                conn1,
                {name: slot.dtype for name, slot in in_slots.items()},
                {name: slot.dtype for name, slot in out_slots.items()},
                deepcopy(params),
                cls.category(),
            ),
            node,
        )

    @classmethod
    def create_standalone(cls) -> "Node":
        """
        Create a new node instance in the current process without a connection to the manager.

        ### Returns
        `Node`
            The node instance.
        """
        # generate arguments for the node
        in_slots, out_slots, params = cls._configure(cls)
        # instantiate the node in the current process
        return cls(None, in_slots, out_slots, params, NodeEnv.STANDALONE)

    @classmethod
    def category(cls) -> str:
        """
        Returns the category of the node, i.e. the name of the node's module.

        ### Returns
        `str`
            The category of the node.
        """
        return cls.__module__.split(".")[-2]

    @property
    def assets_path(self) -> PosixPath:
        """
        Returns the absolute path to the assets folder of goofi-pipe.

        ### Returns
        `PosixPath`
            The path to the assets folder of the node.
        """
        return pkg_resources.files(assets)

    @property
    def data_path(self) -> str:
        """
        Returns the absolute path to the data folder of goofi-pipe.

        ### Returns
        `str`
            The path to the data folder of the node.
        """
        return join(dirname(dirname(dirname(__file__))), "data")

    @property
    @require_init
    def alive(self) -> bool:
        return self._alive

    @property
    @require_init
    def params(self) -> NodeParams:
        return self._params

    @property
    @require_init
    def input_slots(self) -> Dict[str, InputSlot]:
        return self._input_slots

    @property
    @require_init
    def output_slots(self) -> Dict[str, OutputSlot]:
        return self._output_slots

    @staticmethod
    def config_input_slots() -> Dict[str, Union[InputSlot, DataType]]:
        """
        This method is called when the node is instantiated. It should return a dict containing the input slots
        of the node. The keys are the names of the input slots, and the values are either InputSlot instances or
        DataType instances.

        ### Returns
        `Dict[str, Union[InputSlot, DataType]]`
            A dict containing the input slots of the node.
        """
        return {}

    @staticmethod
    def config_output_slots() -> Dict[str, Union[OutputSlot, DataType]]:
        """
        This method is called when the node is instantiated. It should return a dict containing the output slots
        of the node. The keys are the names of the output slots, and the values are either OutputSlot instances or
        DataType instances.

        ### Returns
        `Dict[str, Union[OutputSlot, DataType]]`
            A dict containing the output slots of the node.
        """
        return {}

    @staticmethod
    def config_params() -> Dict[str, Dict[str, Any]]:
        """
        This method is called when the node is instantiated. It should return a dict containing the parameters
        of the node. The keys are the names of the parameter groups, and the values are dicts containing the
        parameters of each group.

        ### Returns
        `Dict[str, Dict[str, Any]]`
            A dict containing the parameters of the node. The values may be any type that is supported by the
            `Param` class.
        """
        return {}

    @require_init
    def setup(self) -> None:
        """
        This method is called after the node is instantiated. It can be used to set up the node.
        """
        pass

    @abstractmethod
    def process(self, **kwargs) -> Dict[str, Tuple[Any, Dict[str, Any]]]:
        """
        This method is called when the node is triggered. It should process the input data and return a dict
        containing the output data.

        ### Parameters
        `**kwargs` : Any
            The input data.

        ### Returns
        `Dict[str, Tuple[Any, Dict[str, Any]]]`
            A dict containing the output data. The keys are the names of the output slots, and the values are
            tuples containing the data and metadata of the output data.
        """
        pass

    def terminate(self) -> None:
        """
        This method is called when the node is terminated. It can be used to clean up any resources used by the
        node.
        """
        pass
